Package: dsTidyverse
Type: Package
Title: 'DataSHIELD' 'Tidyverse' Serverside Package
Version: 1.0.4
Authors@R: 
    c(
      person(given = "Tim",
             family = "Cadman",
             role = c("aut", "cre"),
             email = "t.j.cadman@umcg.nl",
             comment = c(ORCID = "0000-0002-7682-5645")), 
     person(given = "Mariska",
             family = "Slofstra",
             role = c("aut"),
             email = "m.k.slofstra@umcg.nl",
             comment = c(ORCID = "0000-0002-0400-0468")),
     person(given = "Stuart",
            family = "Wheater",
            role = c("aut")),
     person(given = "Demetris",
            family = "Avraam",
            role = c("aut"))
       )
Maintainer: Tim Cadman <t.j.cadman@umcg.nl>
Description: Implementation of selected 'Tidyverse' functions within 'DataSHIELD', an open-source federated analysis solution in R. Currently, DataSHIELD contains very limited tools for data manipulation, so the aim of this package is to improve the researcher experience by implementing essential functions for data manipulation, including subsetting, filtering, grouping, and renaming variables. This is the serverside package which should be installed on the server holding the data, and is used in conjuncture with the clientside package 'dsTidyverseClient' which is installed in the local R environment of the analyst. For more information, see <https://www.tidyverse.org/> and <https://datashield.org/>.
License: LGPL (>= 2.1)
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: rlang, cli
Suggests: testthat (>= 3.0.0), tibble, DSLite, dsBaseClient, dsBase,
        DSI, dplyr, purrr, mockery
Config/testthat/edition: 3
Additional_repositories: https://cran.obiba.org/
NeedsCompilation: no
Packaged: 2025-02-27 09:19:47 UTC; tcadman
Author: Tim Cadman [aut, cre] (<https://orcid.org/0000-0002-7682-5645>),
  Mariska Slofstra [aut] (<https://orcid.org/0000-0002-0400-0468>),
  Stuart Wheater [aut],
  Demetris Avraam [aut]
Repository: CRAN
Date/Publication: 2025-02-27 09:40:06 UTC
Built: R 4.3.3; ; 2025-02-27 10:59:41 UTC; unix

Package: dsTidyverse
Type: Package
Title: 'DataSHIELD' 'Tidyverse' Server-Side Package
Version: 1.2.1
Authors@R: 
    c(
      person(given = "Tim",
             family = "Cadman",
             role = c("aut", "cre"),
             email = "t.j.cadman@umcg.nl",
             comment = c(ORCID = "0000-0002-7682-5645")), 
     person(given = "Mariska",
             family = "Slofstra",
             role = c("aut"),
             email = "m.k.slofstra@umcg.nl",
             comment = c(ORCID = "0000-0002-0400-0468")),
     person(given = "Stuart",
            family = "Wheater",
            role = c("aut")),
     person(given = "Demetris",
            family = "Avraam",
            role = c("aut"))
       )
Maintainer: Tim Cadman <t.j.cadman@umcg.nl>
Description: Implementation of selected 'Tidyverse' functions within 'DataSHIELD', an open-source federated analysis solution in R. Currently, DataSHIELD contains very limited tools for data manipulation, so the aim of this package is to improve the researcher experience by implementing essential functions for data manipulation, including subsetting, filtering, grouping, and renaming variables. This is the server-side package which should be installed on the server holding the data, and is used in conjunction with the client-side package 'dsTidyverseClient' which is installed in the local R environment of the analyst. For more information, see <https://tidyverse.org/> and <https://datashield.org/>.
License: LGPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: rlang, cli, dplyr
Suggests: testthat (>= 3.0.0), tibble, DSLite, dsBaseClient, dsBase,
        DSI, purrr, mockery, knitr, rmarkdown
Config/testthat/edition: 3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-27 12:24:24 UTC; tcadman
Author: Tim Cadman [aut, cre] (ORCID: <https://orcid.org/0000-0002-7682-5645>),
  Mariska Slofstra [aut] (ORCID: <https://orcid.org/0000-0002-0400-0468>),
  Stuart Wheater [aut],
  Demetris Avraam [aut]
Repository: CRAN
Date/Publication: 2026-04-27 12:40:09 UTC
Built: R 4.6.0; ; 2026-04-27 21:41:35 UTC; unix

# Generated by roxygen2: do not edit by hand

export(arrangeDS)
export(asTibbleDS)
export(bindColsDS)
export(bindRowsDS)
export(caseWhenDS)
export(checkPermissivePrivacyControlLevel)
export(distinctDS)
export(filterDS)
export(groupByDS)
export(groupKeysDS)
export(ifElseDS)
export(listPermittedTidyverseFunctionsDS)
export(mutateDS)
export(renameDS)
export(selectDS)
export(sliceDS)
export(ungroupDS)
importFrom(cli,cli_abort)
importFrom(dplyr,desc)
importFrom(rlang,eval_tidy)
importFrom(rlang,parse_expr)

# dsTidyverse
<!-- badges: start -->
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/dsTidyverse)](https://CRAN.R-project.org/package=dsTidyverse)
[![CircleCI](https://circleci.com/gh/molgenis/ds-tidyverse.svg?style=shield)](https://app.circleci.com/pipelines/github/molgenis/ds-tidyverse)
[![codecov](https://codecov.io/gh/molgenis/ds-tidyverse/branch/master/graph/badge.svg?token=ITPMERAWYI)](https://app.codecov.io/gh/molgenis/ds-tidyverse)
<!-- badges: end -->

This is the DataSHIELD implementation of selected [Tidyverse](https://tidyverse.org/) functions. 
Currently most of these functions are from dplyr, however in the future functions from other
Tidyverse packages can be included.

## Installation
- Install dsTidyverse on your armadillo or opal server.
- Whitelist dsTidyverse if required
- Install dsTidyverseClient on your local machine

## Functions/Features
### From the `dplyr` package:
- **`select`**: Choose columns from a data frame.
- **`rename`**: Rename columns in a data frame.
- **`mutate`**: Create or modify columns.
- **`if_else`**: A vectorised conditional function.
- **`case_when`**: A general vectorised conditional function.
- **`bind_cols`**: Combine data frames by columns.
- **`bind_rows`**: Combine data frames by rows.
- **`filter`**: Filter rows based on conditions.
- **`slice`**: Select rows by position.
- **`arrange`**: Arrange rows by values of a column or multiple columns.
- **`group_by`**: Group data by one or more columns.
- **`ungroup`**: Remove grouping from data.
- **`group_keys`**: Retrieve the group keys from a grouped data frame.
- **`distinct`**: Return unique rows based on certain columns.

### From the `tibble` package:
- **`as_tibble`**: Convert data to a tibble.

## Privacy settings with dsTidyverse
Please read the separate article **Privacy settings with dsTidyverse** for detailed instructions on managing privacy within dsTidyverse.

## Contributing
If there are functions in this list you would like implemented, please either attempt to do so 
yourself and submit a pull request, or submit a feature request in the `issues` section. See 
the "CONTRIBUTING" file for more information.

#' @title Bind multiple data frames by column
#' @description DataSHIELD implementation of \code{dplyr::bind_cols}.
#' @param to_combine 	Data frames to combine. Each argument can either be a data frame, a list that
#' could be a data frame, or a list of data frames. Columns are matched by name, and any missing
#' columns will be filled with NA.
#' @param .name_repair One of "unique", "universal", or "check_unique". See
#' \code{vctrs::vec_as_names()} for the meaning of these options.
#' @return A data frame the same type as the first element of \code{to_combine}
#' @export
bindColsDS <- function(to_combine = NULL, .name_repair = NULL) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  to_combine <- .decode_tidy_eval(to_combine, .get_encode_dictionary())
  .check_tidy_disclosure(NULL, to_combine, check_df = F)
  other_args <- .paste_character_args(.name_repair)
  call <- .make_tidyverse_call(.data = NULL, "bind_cols", to_combine, other_args, inc_data = F)
  out <- .execute_with_error_handling("bind_cols", call)
  return(out)
}

#' @title Order the rows of a data frame by the values of selected columns
#' @description DataSHIELD implementation of \code{dplyr::arrange}.
#' @param df.name A data frame, data frame extension (e.g. a tibble), or a lazy data frame (e.g. from dbplyr or dtplyr).
#' @param tidy_expr Variables, or functions of variables. Use \code{desc()} to sort a variable in descending
#' order.
#' @param .by_group If TRUE, will sort first by grouping variable. Applies to grouped data frames
#' only.
#' @return An object of the same type as \code{df.name}, typically a data frame or tibble.
#' @export
arrangeDS <- function(tidy_expr, df.name, .by_group) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  .check_tidy_disclosure(df.name, tidy_expr)
  call <- .make_tidyverse_call(df.name, "arrange", tidy_expr, list(.by_group))
  out <- .execute_with_error_handling("arrange", call)
  return(out)
}

#' @title Keep distinct/unique rows
#' @description DataSHIELD implementation of \code{dplyr::distinct}.
#' @param tidy_expr Optional variables to use when determining uniqueness. If there are multiple rows for
#' a given combination of inputs, only the first row will be preserved. If omitted, will use all
#' variables in the data frame.
#' @param df.name A data frame, data frame extension (e.g. a tibble), or a lazy data frame
#' (e.g. from dbplyr or dtplyr).
#' @param .keep_all If TRUE, keep all variables in df.name If a combination of expr is not distinct,
#' this keeps the first row of values.
#' @return An object of the same type as \code{df.name}, typically a data frame or tibble.
#' @export
distinctDS <- function(tidy_expr, df.name, .keep_all) {
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  other_args <- .paste_character_args(.keep_all)
  call <- .make_tidyverse_call(df.name, "distinct", tidy_expr, other_args)
  out <- .execute_with_error_handling("distinct", call)
  .check_subset_disclosure_risk(eval(parse(text = df.name)), out)
  return(out)
}

#' @title Performs dplyr \code{group_keys}.
#' @description DataSHIELD implementation of \code{dplyr::group_keys}
#' @param tidy_select Unused in this function.
#' @param x a grouped tibble.
#' @return A data frame describing the groups.
#' @export
groupKeysDS <- function(tidy_select, x) {
  dsBase::checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  .check_data_name_length(x, listDisclosureSettingsDS())
  call <- .make_tidyverse_call(x, "group_keys", tidy_expr = NULL, other_args = NULL)
  out <- .execute_with_error_handling("group_keys", call)
  .check_group_keys_disclosure_risk(eval(parse(text = x), envir = parent.frame()), out)
  return(out)
}

#' Check Group Keys Disclosure Risk
#'
#' This internal function checks the disclosure risk by comparing the number of groups in the
#' original dataset with the number of groups in the output dataset. If the number of groups in
#' the output dataset is too close to the number of groups in the original dataset, a disclosure
#' risk is detected and an error is thrown.
#'
#' @param original A data frame representing the original data
#' @param out A data frame representing a summary of the groups
#' @return Silent if no disclosure risk, else error is thrown.
#' @noRd
.check_group_keys_disclosure_risk <- function(original, out) {
  nfilter.levels.density <- .get_disclosure_value("nfilter.levels.density")
  dims <- .get_dimensions(original, out)
  .check_n_groups_compared_with_original(dims, nfilter.levels.density)
}

#' Check Number of Groups Compared with Original Dataset
#'
#' This internal function evaluates the number of groups in a subset against a calculated disclosure
#' threshold. If the number of groups in the subset exceeds the threshold, a disclosure risk is
#' detected and an error is raised to prevent the return of group keys that might compromise data
#' privacy.
#'
#' @param dims A list containing the dimensions of the original and subset datasets. It should
#' include `original` and `subset` elements representing the number of groups in the original and
#' subset datasets, respectively.
#' @param disclosure_value A numeric value representing the disclosure threshold
#' @return Silent if no disclosure risk, else error is thrown.
#' @noRd
.check_n_groups_compared_with_original <- function(dims, disclosure_value) {
  threshold <- dims$original * disclosure_value
  if (threshold < dims$subset) {
    cli_abort(
      c("The group keys cannot be returned due to a disclosure risk",
        "i" = "The length of the dataset is {dims$original} and the number of groups is {dims$subset}",
        "i" = "Based on current disclosure settings the minimum number of groups required to return
      these values must be >= {threshold}",
        call = NULL
      )
    )
  }
}

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>%
  group_by(cyl) %>%
  mutate(drop_test = factor("a", levels = c("a", "b")))
mtcars_bad_group <- mtcars %>% group_by(qsec)

login_data <- .prepare_dslite(
  NULL, "groupKeysDS",
  list(mtcars = mtcars, mtcars_group = mtcars_group, mtcars_bad_group = mtcars_bad_group)
)
conns <- datashield.login(logins = login_data)

datashield.assign.table(conns, "mtcars", "mtcars")
datashield.assign.table(conns, "mtcars_group", "mtcars_group")
datashield.assign.table(conns, "mtcars_bad_group", "mtcars_bad_group")

test_that("groupKeysDS correctly returns keys when no disclosure risk", {
  good_keys_call <- .make_tidyverse_call("mtcars_group", "group_keys", tidy_expr = NULL, other_args = NULL)

  expect_equal(
    eval(good_keys_call),
    tibble(cyl = c(4, 6, 8))
  )
})

test_that("groupKeysDS fails when data doesn't exist", {
  no_data_call <- .make_tidyverse_call("doesnt_exist", "group_keys", tidy_expr = NULL, other_args = NULL)
  expect_error(
    eval(no_data_call),
    "object 'doesnt_exist' not found"
  )
})

test_that("groupKeys passes when called directly and no disclosure risk", {
  skip_if_not_installed("dsBaseClient")
  call_direct <- call("groupKeysDS", NULL, "mtcars_group")
  groups_returned <- datashield.aggregate(conns, call_direct)

  expect_equal(
    class(groups_returned[[1]]),
    c("tbl_df", "tbl", "data.frame")
  )

  expect_equal(
    groups_returned[[1]],
    tibble(cyl = c(4, 6, 8))
  )
})

test_that("groupKeys fails when called directly with disclosure risk", {
  skip_if_not_installed("dsBaseClient")
  call_disc <- call("groupKeysDS", "mtcars_bad_group")
  expect_error(datashield.aggregate(conns, call_disc))
})

test_that(".check_n_groups_compared_with_original doesn't through error if number of groups not too high", {
  density_val <- 0.3
  dims_valid <- list(original = 100, subset = 10)
  expect_silent(.check_n_groups_compared_with_original(dims_valid, density_val))
})

test_that(".check_n_groups_compared_with_original detects disclosure risk correctly", {
  density_val <- 0.3
  dims_high_risk <- list(original = 100, subset = 80)

  expect_error(
    .check_n_groups_compared_with_original(dims_high_risk, 0.3),
    "The group keys cannot be returned due to a disclosure risk"
  )
})

test_that(".check_group_keys_disclosure_risk doesn't through error if number of groups not too high", {
  original_valid <- data.frame(id = 1:100)
  out_valid <- data.frame(id = 1:20)

  expect_silent(
    with_mocked_bindings(
      .check_group_keys_disclosure_risk(original_valid, out_valid),
      ".get_disclosure_value" = function(value) 0.33
    )
  )
})

test_that(".check_group_keys_disclosure_risk throws error if number of groups too high", {
  original_valid <- data.frame(id = 1:100)
  out_valid <- data.frame(id = 1:90)

  expect_error(
    with_mocked_bindings(
      .check_group_keys_disclosure_risk(original_valid, out_valid),
      ".get_disclosure_value" = function(value) 0.33
    )
  )
})

#' @title Performs dplyr case_when
#' @description DataSHIELD implementation of \code{dplyr::case_when}.
#' @param tidy_expr A sequence of two-sided formulas. The left hand side (LHS) determines which
#' values match this case. The right hand side (RHS) provides the replacement value.
#' The LHS inputs must evaluate to logical vectors.
#' The RHS inputs will be coerced to their common type.
#' All inputs will be recycled to their common size. That said, we encourage all LHS inputs to be
#' the same size. Recycling is mainly useful for RHS inputs, where you might supply a size 1 input
#' that will be recycled to the size of the LHS inputs.
#' NULL inputs are ignored.
#' @param .default The value used when all of the LHS inputs return either FALSE or NA.
#' @param .ptype An optional prototype declaring the desired output type. If supplied, this overrides the common type of true, false, and missing.
#' @param .size An optional size declaring the desired output size. If supplied, this overrides the size of condition.
#' @return A vector with the same size as the common size computed from the inputs in \code{tidy_expr} and the same type as the common type of the RHS inputs in \code{tidy_expr}.
#' @export
caseWhenDS <- function(tidy_expr = NULL, .default = NULL, .ptype = NULL, .size = NULL) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  .check_tidy_disclosure(NULL, tidy_expr, check_df = F)
  other_args <- .paste_character_args(.default, .ptype, .size)
  call <- .make_tidyverse_call(.data = NULL, "case_when", tidy_expr, other_args, inc_data = F)
  out <- .execute_with_error_handling("case_when", call)
  return(out)
}

#' Generate an encoding key which is used for encoding and decoding strings to pass the R parser
#'
#' @return A list containing the encoding key, with 'input' specifying the characters to be encoded
#' and 'output' specifying their corresponding encoded values.
#' @noRd
.get_encode_dictionary <- function() {
  encode_list <- list(
    input = c("(", ")", "\"", ",", " ", "!", "&", "|", "'", "=", "+", "-", "*", "/", "^", ">", "<", "~", "\n", "%"),
    output = c(
      "$LB$", "$RB$", "$QUOTE$", "$COMMA$", "$SPACE$", "$EXCL$", "$AND$", "$OR$",
      "$APO$", "$EQU$", "$ADD$", "$SUB$", "$MULT$", "$DIVIDE$", "$POWER$", "$GT$", "$LT$", "$TILDE$", "$LINE$", "$PCT$"
    )
  )
  return(encode_list)
}

#' Decode a string using the provided encoding key.
#'
#' @param input_string The encoded string passed through the R parser.
#' @param encode_key The encoding key generated by '.get_encode_dictionary()'.
#' @return The decoded string.
#' @noRd
.decode_tidy_eval <- function(input_string, encode_key) {
  encode_vec <- encode_key$input
  names(encode_vec) <- encode_key$output
  output_string <- Reduce(function(out, pattern) {
    gsub(pattern, encode_vec[[pattern]], out, fixed = TRUE)
  }, names(encode_vec), input_string)
  return(output_string)
}

#' Create a tidyverse call expression
#'
#' This function creates a tidyverse call expression in string form,
#' using the provided data frame, function name, tidy select statement,
#' and additional arguments.
#'
#' @param .data The name of the data frame.
#' @param fun The name of the function to be applied (e.g., "select").
#' @param tidy_select The tidy select statement (e.g., column names).
#' @param other_args Additional arguments to the function.
#' @param inc_data Boolean, whether to include a data object in the call
#' @importFrom rlang parse_expr
#' @return An expression object of the tidyverse call.
#' @noRd
.make_tidyverse_call <- function(.data, fun, tidy_expr, other_args = NULL, inc_data = TRUE) {
  if (!is.null(tidy_expr) & is.null(other_args)) {
    tidy_string <- paste0("dplyr::", fun, "(", tidy_expr, ")")
  } else if (!length(tidy_expr) == 0 & !is.null(other_args)) {
    tidy_string <- paste0("dplyr::", fun, "(", tidy_expr, ", ", other_args, ")")
  } else if (length(tidy_expr) == 0 & !is.null(other_args)) {
    tidy_string <- paste0("dplyr::", fun, "(", other_args, ")")
  } else if (length(tidy_expr) == 0 & is.null(other_args)) {
    tidy_string <- paste0("dplyr::", fun, "(", ")")
  }

  if (inc_data) {
    tidy_string <- paste0(.data, " |> ", tidy_string)
  }
  return(rlang::parse_expr(tidy_string))
}

#' Paste Character Arguments
#'
#' This function takes any number of arguments and returns a formatted string
#' that includes the argument names and their corresponding values.
#'
#' @param ... Any number of arguments.
#' @return A character string with the argument names and values.
#' @noRd
.paste_character_args <- function(...) {
  arg_values <- lapply(list(...), deparse)
  call_stack <- sys.call()
  arg_names <- as.character(call_stack)[-1]
  names(arg_values) <- arg_names
  args_formatted <- mapply(function(name, value) paste0(name, " = ", value),
                           arg_names, arg_values, SIMPLIFY = FALSE)
  args_as_vector <- paste(unlist(args_formatted), collapse = ", ")
  return(args_as_vector)
}


#' Build an evaluation environment with permitted dplyr functions
#'
#' Creates a new environment parented by \code{caller_env} and populates it with
#' permitted dplyr functions so they are available during tidy evaluation.
#'
#' @param caller_env The environment containing the data objects (e.g. the DataSHIELD server environment).
#' @return A new environment with permitted dplyr functions, parented by \code{caller_env}.
#' @importFrom dplyr desc
#' @noRd
.build_eval_env <- function(caller_env) {
  eval_env <- new.env(parent = caller_env)
  permitted <- listPermittedTidyverseFunctionsDS()
  dplyr_ns <- asNamespace("dplyr")
  for (fn in permitted) {
    if (exists(fn, envir = dplyr_ns, inherits = FALSE)) {
      eval_env[[fn]] <- get(fn, envir = dplyr_ns)
    }
  }
  eval_env
}

#' Evaluate an expression and handle errors gracefully.
#'
#' @param fun The name of the calling function, used in error messages.
#' @param string_as_expr An expression to be evaluated.
#' @importFrom cli cli_abort
#' @importFrom rlang eval_tidy
#' @return The result of evaluating the expression, or an error message if evaluation fails.
#' @noRd
.execute_with_error_handling <- function(fun, string_as_expr) {
  caller_env <- parent.frame(2)
  eval_env <- .build_eval_env(caller_env)
  object_out <- tryCatch(
    eval_tidy(string_as_expr, env = eval_env),
    error = function(e) {
      cli_abort(
        c("`{fun}DS` returned the following error:", "i" = conditionMessage(e)),
        call = NULL
      )
    }
  )
  return(object_out)
}

#' @title listDisclosureSettingsDS
#' @description This server-side function is an aggregate function that is called by the
#' ds.listDisclosureSettings
#' @details For more details see the extensive header for ds.listDisclosureSettings
#' @author Paul Burton, Demetris Avraam for DataSHIELD Development Team
#' @noRd
listDisclosureSettingsDS <- function() {
  ds.privacyControlLevel <- getOption("datashield.privacyControlLevel")
  if (is.null(ds.privacyControlLevel)) {
    ds.privacyControlLevel <- getOption("default.datashield.privacyControlLevel")
  }

  nf.tab <- getOption("nfilter.tab")
  if (is.null(nf.tab)) {
    nf.tab <- getOption("default.nfilter.tab")
  }
  nf.subset <- getOption("nfilter.subset")
  if (is.null(nf.subset)) {
    nf.subset <- getOption("default.nfilter.subset")
  }
  nf.glm <- getOption("nfilter.glm")
  if (is.null(nf.glm)) {
    nf.glm <- getOption("default.nfilter.glm")
  }
  nf.string <- getOption("nfilter.string")
  if (is.null(nf.string)) {
    nf.string <- getOption("default.nfilter.string")
  }
  nf.stringShort <- getOption("nfilter.stringShort")
  if (is.null(nf.stringShort)) {
    nf.stringShort <- getOption("default.nfilter.stringShort")
  }
  nf.kNN <- getOption("nfilter.kNN")
  if (is.null(nf.kNN)) {
    nf.kNN <- getOption("default.nfilter.kNN")
  }
  nf.levels.density <- getOption("nfilter.levels.density")
  if (is.null(nf.levels.density)) {
    nf.levels.density <- getOption("default.nfilter.levels.density")
  }
  nf.levels.max <- getOption("nfilter.levels.max")
  if (is.null(nf.levels.max)) {
    nf.levels.max <- getOption("default.nfilter.levels.max")
  }
  nf.noise <- getOption("nfilter.noise")
  if (is.null(nf.noise)) {
    nf.noise <- getOption("default.nfilter.noise")
  }
  nfilter.privacy.old <- getOption("datashield.privacyLevel")

  return(list(
    datashield.privacyControlLevel = ds.privacyControlLevel, nfilter.tab = nf.tab, nfilter.subset = nf.subset,
    nfilter.glm = nf.glm, nfilter.string = nf.string,
    nfilter.stringShort = nf.stringShort, nfilter.kNN = nf.kNN, nfilter.levels.density = nf.levels.density,
    nfilter.levels.max = nf.levels.max, nfilter.noise = nf.noise, nfilter.privacy.old = nfilter.privacy.old
  ))
}

#' List of Permitted Tidyverse Functions
#'
#' This function returns a vector of function names that are permitted to be passed within the
#' dsTidyverse functions, e.g. within the `tidy_select` argument of `ds.mutate.`
#' Additional functions can be added via the \code{tidyverse.permitted.functions} R option.
#' Only functions from base R, dplyr, or tidyselect are supported.
#'
#' @return A character vector of function names, each representing a permitted function. Functions
#' not included in this list will be blocked.
#' @export
listPermittedTidyverseFunctionsDS <- function() {
  defaultFunctions <- c(
    "everything", "last_col", "group_cols", "starts_with", "ends_with", "contains",
    "matches", "num_range", "all_of", "any_of", "where", "rename", "mutate", "if_else",
    "case_when", "mean", "median", "desc",
    "exp", "sqrt", "scale", "round", "floor", "ceiling", "abs", "sd", "var",
    "sin", "cos", "tan", "asin", "acos", "atan", "c", "as.character", "as.integer", "as.numeric",
    "lag", "cumsum", "is.na", "as.Date"
  )

  permittedFunctions <- getOption("tidyverse.permitted.functions", defaultFunctions)

  return(permittedFunctions)
}

#' Check Subset Disclosure Risk
#'
#' This function checks the disclosure risk when applying creating a subset of a dataset.
#' It evaluates the subset size and the difference in rows between the original
#' and subsetted data to ensure they meet the minimum threshold specified by
#' `nfilter.subset`.
#'
#' @param .data A string representing the name of the original dataset.
#' @param out The subset.
#' @keywords internal
#' @return None. The function will throw an error if disclosure risk is detected.
#' @noRd
.check_subset_disclosure_risk <- function(original, out) {
  nfilter.subset <- .get_nfilter_subset_value()
  dims <- .get_dimensions(original, out)
  .check_subset_size(dims$subset, nfilter.subset)
  .check_rows_compared_with_original(dims$original, dims$subset, nfilter.subset)
}

#' Get `nfilter.subset` Value
#'
#' This function retrieves the value of `nfilter.subset` from the disclosure
#' settings.
#'
#' @keywords internal
#' @return The value of `nfilter.subset` from the disclosure settings.
#' @noRd
.get_nfilter_subset_value <- function() {
  return(
    listDisclosureSettingsDS()$nfilter.subset
  )
}

#' Get Dimensions of Original and Subset Data
#'
#' This function calculates the number of rows in the original and subsetted datasets.
#'
#' @param .data A string representing the name of the original dataset.
#' @param out The filtered dataset object.
#' @keywords internal
#' @return A list containing the number of rows in the original and subsetted datasets.
#' @noRd
.get_dimensions <- function(original, out) {
  return(
    list(
      original = dim(original)[[1]],
      subset = dim(out)[[1]]
    )
  )
}

#' Check Subset Size
#'
#' This function checks if the number of rows in the subsetted data is below the
#' threshold defined by `nfilter.subset`. If it is, the function throws an error.
#'
#' @param subset_rows The number of rows in the subsetted data.
#' @param nfilter.subset The minimum allowed size for a subset as defined in the
#' disclosure settings.
#' @keywords internal
#' @return None. The function will throw an error if the subset size is too small.
#' @noRd
.check_subset_size <- function(subset_rows, nfilter.subset) {
  if (subset_rows < nfilter.subset) {
    cli_abort(
      "Subset to be created is too small (< nfilter.subset)",
      call = NULL
    )
  }
}

#' Check Rows Compared with Original
#'
#' This function checks the difference in the number of rows between the original
#' and subsetted data. If the difference is smaller than `nfilter.subset` but
#' greater than zero, it raises an error, indicating a potential disclosure risk.
#'
#' @param original_rows The number of rows in the original dataset.
#' @param subset_rows The number of rows in the subsetted dataset.
#' @param nfilter.subset The minimum allowed difference between the original and
#' subsetted data, as defined in the disclosure settings.
#' @keywords internal
#' @return None. The function will throw an error if a potential disclosure risk
#' is detected.
#' @noRd
.check_rows_compared_with_original <- function(original_rows, subset_rows, nfilter.subset) {
  diff <- original_rows - subset_rows
  if ((diff < nfilter.subset) & (diff > 0)) {
    cli_abort(
      "The difference in row length between the original dataframe and the new dataframe is {diff},
      which is lower than the value of nfilter.subset ({nfilter.subset}). This could indicate a potential subsetting
      attack which will be recorded in the server-side logs. Please review the filter expression.",
      call = NULL
    )
  }
}

#' Get single disclosure setting
#'
#' @param setting the disclosure setting to return
#' @return The value of setting from the disclosure settings.
#' @noRd
.get_disclosure_value <- function(setting) {
  return(
    listDisclosureSettingsDS()[[setting]]
  )
}

#' Check if the functions used in tidy evaluation are permitted.
#'
#' @param args_as_string The string representation of the arguments.
#' @importFrom cli cli_abort
#' @details To avoid users attempting to maliciously pass functions to the servervia the tidy_select
#' argument, a regex is used to first extract all functions from the string by identifying any characters
#' with the format 'name('. These are then checked against permitted arguments, which are selection
#' helpers described in the ?select documentation.#'
#' @noRd
.check_function_names <- function(args_as_string) {
  permitted_tidy_select <- listPermittedTidyverseFunctionsDS()
  matches <- gregexpr("[A-Za-z][A-Za-z0-9_]*(?:\\.[A-Za-z0-9_]+)*(?=\\()", args_as_string, perl = TRUE)
  function_names <- unlist(regmatches(args_as_string, matches))
  any_banned_functions <- function_names[!function_names %in% permitted_tidy_select]
  if (length(any_banned_functions) > 0) {
    message <- c(
      "Values passed to `expr` may only contain permitted functions.",
      "Permitted functions are {permitted_tidy_select}.",
      "`{any_banned_functions}` {?is/are} not {?a/} permitted function{?s}."
    )
    names(message) <- c("", "i", "i")
    cli_abort(message, call = NULL)
  }
}

#' Check if the length of variable names in tidy evaluation exceeds a nfilter.string threshold.
#'
#' @param args_as_string The string representation of the arguments.
#' @param disclosure list of disclosure settings, length of number of cohorts
#' @importFrom cli cli_abort
#' @details To check users are not passing variable names which are too long, first a regex extracts
#' variable names from the list passed to `tidy_select`. It then checks the lengths of these against
#' the value passed to nfilter.string.#'
#' @noRd
.check_variable_length <- function(args_as_string, disclosure) {
  matches <- gregexpr("\\b\\w+\\b(?!\\()", args_as_string, perl = TRUE)
  variable_names <- unlist(regmatches(args_as_string, matches))

  if(length(variable_names) > 0) {
    variable_lengths <- sapply(variable_names, nchar)
    over_filter_thresh <- .check_exceeds_threshold(variable_names, variable_lengths, disclosure$nfilter.string)

  if (length(over_filter_thresh) > 0) {
    disclosure_message <- .format_disclosure_errors(disclosure)
    cli_abort(disclosure_message, call = NULL)
  }
  }
}

#' Check Which Variables Exceed a Threshold
#'
#' This function checks if the lengths of the given variables exceed a specified threshold.
#' It returns the names of the variables whose lengths are greater than the threshold.
#'
#' @param variable_names A character vector containing the names of the variables.
#' @param variable_lengths A numeric vector containing the lengths of the variables, in the same order as `variable_names`.
#' @param threshold A numeric value specifying the threshold that each variable length is compared against.
#' @return A character vector of variable names where the corresponding lengths exceed the threshold.
#' @noRd
.check_exceeds_threshold <- function(variable_names, variable_lengths, threshold) {
  return(
    variable_names[sapply(variable_lengths, function(x) x > threshold)]
  )
}

#' Format Errors
#'
#' Format errors into a character vector with specified prefix.
#'
#' This function formats a list of errors into a character vector with each
#' error message prefixed by a cross.
#'
#' @param errors A list of errors to be formatted.
#' @param disclosure list of disclosure settings, length of number of cohorts
#' @return A character vector containing formatted error messages.
#' @noRd
.format_disclosure_errors <- function(disclosure) {
  out <- c(
    "Error: The maximum length of columns specified in `tidy_select` must be shorter than
           nfilter.string. ", "The values of nfilter.string are: ", disclosure$nfilter.string,
    "{over_filter_thresh} {?is/are} longer than this: "
  )
  names(out) <- c("", "i", "", "i")
  return(out)
}

#' Check tidyverse disclosure Settings
#'
#' @param df.name A character string specifying the name of the dataframe.
#' @param tidy_select A tidy selection specification of columns.
#' @return None. This function is used for its side effects of checking disclosure settings.
#' @noRd
.check_tidy_disclosure <- function(df.name, tidyselect, check_df = TRUE) {
  disc_settings <- listDisclosureSettingsDS()
  if (check_df) {
    .check_data_name_length(df.name, disc_settings)
  }
  .check_function_names(tidyselect)
  .check_variable_length(tidyselect, disc_settings)
}

#' Check that the length of the character string provided to `.data` does not exceed the value of
#' nfilter.string
#'
#' @param .data The data object to be analyzed.
#' @param disclosure list of disclosure settings, length of number of cohorts
#' @noRd
.check_data_name_length <- function(.data, disclosure) {
  data_length <- nchar(.data)
  if (data_length > disclosure$nfilter.string) {
    length_message <- .format_data_length_errors(data_length, disclosure)
    cli_abort(length_message)
  }
}

#' Format errors for `.check_data_name_length`
#'
#' @param data_length The length of the argument to `data`.
#' @param disclosure list of disclosure settings, length of number of cohorts
#' @return A character vector containing the error message.
#' @details The function constructs an error message indicating the length of the string passed to
#' `.data` and the allowed length specified in the `disclosure` argument.
#' @noRd
.format_data_length_errors <- function(data_length, disclosure) {
  out <- c(
    "Error: The length of string passed to `df.name` must be less than nfilter.string.",
    "The value of nfilter.string is: ", disclosure$nfilter.string,
    "The length of `df.name` is: ", data_length
  )
  names(out) <- c("", "i", "", "i", "")
  return(out)
}

#' Check Filter Disclosure Risk
#'
#' This function checks the disclosure risk when applying a filter on a dataset.
#' It evaluates the subset size and the difference in rows between the original
#' and subsetted data to ensure they meet the minimum threshold specified by
#' `nfilter.subset`.
#'
#' @param .data A string representing the name of the original dataset.
#' @param out The filtered dataset object.
#' @keywords internal
#' @return None. The function will throw an error if disclosure risk is detected.
#' @noRd
.check_subset_disclosure_risk <- function(original, out) {
  nfilter.subset <- .get_nfilter_subset_value()
  dims <- .get_dimensions(original, out)
  .check_subset_size(dims$subset, nfilter.subset)
  .check_rows_compared_with_original(dims$original, dims$subset, nfilter.subset)
}

#'
#' @title checkPermissivePrivacyControlLevel
#' @description This server-side function check that the server is running in "permissive" privacy control level.
#' @details Tests whether the R option "datashield.privacyControlLevel" is set to "permissive", if it isn't
#' will cause a call to stop() with the message "BLOCKED: The server is running in 'non-permissive' mode which
#' has caused this method to be blocked".
#' @param privacyControlLevels is a vector of strings which contains the privacy control level names which are permitted by the calling method.
#' @importFrom cli cli_abort
#' @author Wheater, Dr SM., DataSHIELD Team.
#' @return Returns an error if the method is not permitted; otherwise, no value is returned.
#' @export
checkPermissivePrivacyControlLevel <- function(privacyControlLevels){
  disclosureSettings <- listDisclosureSettingsDS()
  if (is.null(disclosureSettings) || is.null(disclosureSettings$datashield.privacyControlLevel) ||
      (! any(disclosureSettings$datashield.privacyControlLevel %in% privacyControlLevels))) {
    cli_abort("BLOCKED: The server is running in 'non-permissive' mode which has caused this method
              to be blocked", call. = TRUE)
  }

  invisible()
}

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
login_data <- .prepare_dslite("filterDS", NULL, list(mtcars = mtcars, mtcars_group = mtcars_group))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

options(
  datashield.privacyControlLevel = "banana",
  nfilter.tab = 3,
  nfilter.subset = 3,
  nfilter.glm = 0.33,
  nfilter.string = 80,
  nfilter.stringShort = 20,
  nfilter.kNN = 3,
  nfilter.levels.density = 0.33,
  nfilter.levels.max = 40,
  nfilter.noise = 0.25,
  nfilter.privacy.old = 5
)

disc_settings <- listDisclosureSettingsDS()

test_that(".make_tidyverse_call creates call with no additional arguments", {
  input_string <- "asd, qwe, starts_with('test')"
  expected_string <- rlang::parse_expr('test |> dplyr::select(asd, qwe, starts_with("test"))')
  observed_string <- .make_tidyverse_call(.data = "test", fun = "select", tidy_expr = input_string)
  expect_equal(expected_string, observed_string)
})

extra_args <- c(".keep = \"all\", .before = NULL, .after = \"disp\"")
input_string <- "asd, qwe, starts_with('test')"

test_that(".make_tidyverse_call creates call with additional arguments", {
  expected_string <- rlang::parse_expr('test |> dplyr::select(asd, qwe, starts_with("test"), .keep = "all",
    .before = NULL, .after = "disp")')
  observed_string <- .make_tidyverse_call(.data = "test", fun = "select", tidy_expr = input_string, other_args = extra_args)
  expect_equal(expected_string, observed_string)
})

test_that(".make_tidyverse_call creates call when inc_data = F", {
  expected_string <- rlang::parse_expr('dplyr::select(asd, qwe, starts_with("test"), .keep = "all",
  .before = NULL, .after = "disp")')
  observed_string <- .make_tidyverse_call(.data = "test", fun = "select", tidy_expr = input_string, other_args = extra_args, inc_data = F)
  expect_equal(expected_string, observed_string)
})

test_that(".make_tidyverse_call creates call when inc_data = F", {
  expected_string <- rlang::parse_expr('dplyr::select(asd, qwe, starts_with("test"), .keep = "all",
  .before = NULL, .after = "disp")')
  observed_string <- .make_tidyverse_call(.data = "test", fun = "select", tidy_expr = input_string, other_args = extra_args, inc_data = F)
  expect_equal(expected_string, observed_string)
})

test_that(".make_tidyverse_call creates call when inc_data = F", {
  expected_string <- rlang::parse_expr('dplyr::select(asd, qwe, starts_with("test"), .keep = "all",
  .before = NULL, .after = "disp")')
  observed_string <- .make_tidyverse_call(.data = "test", fun = "select", tidy_expr = input_string, other_args = extra_args, inc_data = F)
  expect_equal(expected_string, observed_string)
})

mtcars_good_arg <- "mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_good_str <- .make_tidyverse_call(.data = "mtcars", fun = "select", tidy_expr = mtcars_good_arg)

test_that(".execute_with_error_handling works where data and object exists", {
  observed <- .execute_with_error_handling("select", mtcars_good_str)
  expect_equal(colnames(observed), c("mpg", "cyl", "gear", "carb"))
})

mtcars_wrong_data_str <- .make_tidyverse_call(.data = "data_not_here", fun = "select", tidy_expr = mtcars_good_arg)

test_that(".execute_with_error_handling fails with correct message if object doesn't exist", {
  expect_error(
    .execute_with_error_handling("select", mtcars_wrong_data_str),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'mtcars_wrong_data_expr'\\s+not\\s+found"
  )
})

mtcars_missing_col_arg <- "all_of('test_col'), mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_missing_col_str <- .make_tidyverse_call(.data = "mtcars", fun = "select", tidy_expr = mtcars_missing_col_arg)

test_that(".execute_with_error_handling fails with correct message if column doesn't exist", {
  expect_error(
    .execute_with_error_handling("select", mtcars_missing_col_str),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'mtcars_missing_col_expr'\\s+not\\s+found"
  )
})

mtcars_random_arg <- "filter('mpg'), mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_random_str <- .make_tidyverse_call(.data = "mtcars", fun = "select", tidy_expr = mtcars_random_arg)

test_that(".execute_with_error_handling fails with correct message when unrecognised function passed", {
  expect_error(
    .execute_with_error_handling("select", mtcars_random_str),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'mtcars_random_expr'\\s+not\\s+found"
  )
})

mtcars_good_arg <- "mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_good_expr <- .make_tidyverse_call(.data = "mtcars", fun = "select", tidy_expr = mtcars_good_arg)
# mtcars_good_expr <- rlang::parse_expr(mtcars_good_str)

test_that(".tidy_eval_handle_errors works where data and object exists", {
  observed <- .execute_with_error_handling("select", mtcars_good_expr)
  expect_equal(colnames(observed), c("mpg", "cyl", "gear", "carb"))
})

mtcars_good_arg <- "mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_wrong_data_expr <- .make_tidyverse_call(.data = "data_not_here", fun = "select", tidy_expr = mtcars_good_arg)

test_that(".tidy_eval_handle_errors fails with correct message if object doesn't exist", {
  expect_error(
    .execute_with_error_handling("select", mtcars_wrong_data_expr),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'mtcars_wrong_data_expr'\\s+not\\s+found"
  )
})

mtcars_missing_col_arg <- "all_of('test_col'), mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_missing_col_expr <- .make_tidyverse_call(.data = "mtcars", fun = "select", tidy_expr = mtcars_missing_col_arg)

test_that(".tidy_eval_handle_errors fails with correct message if column doesn't exist", {
  expect_error(
    .execute_with_error_handling("select", mtcars_missing_col_expr),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'mtcars_missing_col_expr'\\s+not\\s+found"
  )
})

mtcars_random_arg <- "filter('mpg'), mpg, cyl, starts_with('g'), ends_with('b')"
mtcars_random_expr <- .make_tidyverse_call(.data = "mtcars", fun = "select", tidy_expr = mtcars_random_arg)

test_that(".tidy_eval_handle_errors fails with correct message when unrecognised function passed", {
  expect_error(
    .execute_with_error_handling("select", mtcars_random_expr),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'mtcars_random_expr'\\s+not\\s+found"
  )
})

test_that(".get_encode_dictionary returns the expected encoding key", {
  expected_encode_list <- list(
    input = c("(", ")", "\"", ",", " ", "!", "&", "|", "'", "=", "+", "-", "*", "/", "^", ">", "<", "~", "\n", "%"),
    output = c(
      "$LB$", "$RB$", "$QUOTE$", "$COMMA$", "$SPACE$", "$EXCL$", "$AND$", "$OR$", "$APO$", "$EQU$", "$ADD$", "$SUB$", "$MULT$",
      "$DIVIDE$", "$POWER$", "$GT$", "$LT$", "$TILDE$", "$LINE$", "$PCT$"
    )
  )
  actual_encode_list <- .get_encode_dictionary()
  expect_equal(actual_encode_list, expected_encode_list)
})

test_that(".get_encode_dictionary returns the expected encoding key", {
  expected_encode_list <- list(
    input = c("list", "(", ")", "\"", ",", " ", "c", ":", "!", "&", "|", "=="),
    output = c("$LIST$", "$LB$", "$RB$", "$QUOTE$", "$COMMA$", "$SPACE$", "$C$", "$COLON$", "$EXCL$", "$AND$", "$OR$", "$EQUALS$")
  )
  actual_encode_list <- .get_encode_dictionary()
  expect_false(isTRUE(all.equal(expected_encode_list, actual_encode_list)))
})

expected_string <- "asd, qwe, starts_with('test')"

test_that(".decode_tidy_eval correctly decodes an encoded string passed via the R parser", {
  encoded_string <- "asd$COMMA$$SPACE$qwe$COMMA$$SPACE$starts_with$LB$'test'$RB$"
  decoded_string <- .decode_tidy_eval(encoded_string, .get_encode_dictionary())
  expect_equal(decoded_string, expected_string)
})

test_that(".execute_with_error_handling passes if additional arguments are included", {
  mutate_extra_args <- .make_tidyverse_call("mtcars", "mutate", "new_name_1 = mpg, new_name_2 = drat", ".keep = 'all', .before = NULL, .after = \"disp\"")
  observed <- .execute_with_error_handling("mutate", mutate_extra_args)
  expect_equal(colnames(observed), c("mpg", "cyl", "disp", "new_name_1", "new_name_2", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb"))
})

test_that(".execute_with_error_handling passes if all additional arguments are NULL", {
  mutate_null_args <- .make_tidyverse_call("mtcars", "mutate", "new_name_1 = mpg, new_name_2 = drat", ".keep = 'all', .before = NULL, .after = NULL")
  observed <- .execute_with_error_handling("mutate", mutate_null_args)
  expect_equal(colnames(observed), c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "new_name_1", "new_name_2"))
})

test_that(".paste_character_args creates correct string", {
  true <- "high"
  false <- "low"
  missing <- NULL
  ptype <- NULL
  size <- NULL

  expected <- "true = \"high\", false = \"low\", missing = NULL, ptype = NULL, size = NULL"
  expect_equal(
    .paste_character_args(true, false, missing, ptype, size),
    expected
  )
})

test_that(".get_nfilter_subset_value retrieves value correctly", {
  expect_equal(
    .get_nfilter_subset_value(),
    3
  )
})

test_that(".get_dimensions correctly returns dimensions", {
  original <- tibble(a = 1:3, b = 4:6)
  out <- tibble(a = 1:2, b = 1:2)

  returned <- .get_dimensions(original, out)

  expect_equal(
    class(returned),
    "list"
  )

  expect_equal(
    returned,
    list(original = 3, subset = 2)
  )
})

test_that(".check_subset_size returns an error if subset is too small", {
  expect_error(
    .check_subset_size(subset_rows = 3, nfilter.subset = 10),
    ".*Subset to be created is too small \\(< nfilter\\.subset\\).*"
  )
})

test_that(".check_subset_size doesn't return an error if subset is large enough", {
  expect_silent(
    .check_subset_size(subset_rows = 20, nfilter.subset = 2)
  )
})

test_that(".check_rows_compared_with_original returns an error correctly", {
  expect_error(
    .check_rows_compared_with_original(10, 9, 3),
    "The difference in row length between the original dataframe and the new dataframe"
  )
})

test_that(".check_rows_compared_with_original doesn't return an error if different in dimensions is large enough", {
  expect_silent(
    .check_rows_compared_with_original(100, 80, 3)
  )
})

test_that(".check_subset_disclosure_risk returns an error correctly", {
  original <- tibble(a = 1:3, b = 4:6)
  out <- tibble(a = 1:2, b = 1:2)

  expect_error(
    .check_subset_disclosure_risk(original, out),
    ".*Subset to be created is too small \\(< nfilter\\.subset\\).*"
  )

  original <- tibble(a = 1:10, b = 1:10)
  out <- tibble(a = 1:9, b = 1:9)

  expect_error(
    .check_subset_disclosure_risk(original, out),
    "The difference in row length between the original dataframe and the new dataframe"
  )
})

test_that(".check_subset_disclosure_risk doesn't return errors if subset sizes are ok", {
  original <- tibble(a = 1:10, b = 1:10)
  out <- tibble(a = 1:4, b = 1:4)

  expect_silent(
    .check_subset_disclosure_risk(original, out)
  )

  original <- tibble(a = 1:10, b = 1:10)
  out <- tibble(a = 1:5, b = 1:5)

  expect_silent(
    .check_subset_disclosure_risk(original, out)
  )
})

test_that(".check_data_name_length throws an error if length of .data exceeds nfilter.string", {
  .data <- paste(rep("a", 101), collapse = "")
  expect_snapshot(.check_data_name_length(.data, disc_settings), error = TRUE)
})

test_that(".check_data_name_length does not throw an error if length of .data is within nfilter.string", {
  .data <- paste(rep("a", 79), collapse = "")
  expect_silent(.check_data_name_length(.data, disc_settings))
})

arg_permitted <- "asd, sdf, dfg, everything(), starts_with(\"A\"), ends_with(\"Z\")"
arg_unpermitted <- "asd, sdf, dfg, everything(), filter(test == 2), slice(3), mutate(new_name = old_name), starts_with(\"A\"), ends_with(\"Z\")"
small_var <- paste(rep("a", 5), collapse = "")
large_var <- paste(rep("a", 200), collapse = "")

test_that(".check_function_names allows permitted names to pass", {
  expect_silent(.check_function_names(arg_permitted))
})

test_that(".check_function_names blocks unpermitted function names", {
  arg_unpermitted
  expect_snapshot(
    .check_function_names(arg_unpermitted),
    error = TRUE
  )
})

test_that(".check_variable_length allows variables with value less than nfilter.string", {
  expect_silent(
    .check_variable_length(small_var, disc_settings)
  )
})

test_that(".check_variable_length blocks variables with value greater than than nfilter.string", {
  expect_snapshot(
    .check_variable_length(large_var, disc_settings),
    error = TRUE
  )
})

test_that(".tidy_disclosure_checks allows permitted arg to pass", {
  expect_silent(.check_tidy_disclosure("dataframe", arg_permitted))
})

test_that(".tidy_disclosure_checks blocks argument with unpermitted variable length", {
  arg_unpermitted_2 <- paste0(large_var, arg_permitted)
  expect_snapshot(
    .check_tidy_disclosure("dataframe", arg_unpermitted_2),
    error = TRUE
  )
})

test_that(".tidy_disclosure_checks blocks argument with multiple unpermitted function names", {
  arg_unpermitted_3 <- arg_unpermitted
  expect_snapshot(
    .check_tidy_disclosure("dataset", arg_unpermitted_3),
    error = TRUE
  )
})

test_that(".tidy_disclosure_checks blocks argument with single unpermitted function name", {
  arg_unpermitted_4 <- "asd, sdf, dfg, everything(), slice(3), mutate(new_name = old_name), starts_with(\"A\"), ends_with(\"Z\")"
  expect_snapshot(
    .check_tidy_disclosure("dataset", arg_unpermitted_4),
    error = TRUE
  )
})

test_that("checkPermissivePrivacyControlLevel blocks certain functions when not in permissive mode", {
  cally <- call("filterDS", "carb$SPACE$$EQU$$EQU$$SPACE$4", "mtcars", NULL, FALSE)

  options(datashield.privacyControlLevel =  "banana")
  expect_no_error(
    datashield.assign(conns, "test", cally)
  )

  options(datashield.privacyControlLevel =  "non-permissive")
  expect_error(
    datashield.assign(conns, "test", cally)
  )

  options(datashield.privacyControlLevel = "NULL")
  expect_error(
    datashield.assign(conns, "test", cally)
  )

  expect_error(
    testthat::with_mocked_bindings(
      datashield.assign(conns, "test", cally),
      listDisclosureSettingsDS = function() NULL
      )
  )

  expect_error(
    testthat::with_mocked_bindings(
      datashield.assign(conns, "test", cally),
      listDisclosureSettingsDS = function() list(datashield.privacyControlLevel = "NULL")
    )
  )

})

test_that(".listPermittedTidyverseFunctionsDS returns correct default function list ", {

  expect_equal(
    listPermittedTidyverseFunctionsDS(),
    c(
      "everything", "last_col", "group_cols", "starts_with", "ends_with", "contains",
      "matches", "num_range", "all_of", "any_of", "where", "rename", "mutate", "if_else",
      "case_when", "mean", "median", "desc",
      "exp", "sqrt", "scale", "round", "floor", "ceiling", "abs", "sd", "var",
      "sin", "cos", "tan", "asin", "acos", "atan", "c", "as.character", "as.integer", "as.numeric",
      "lag", "cumsum", "is.na", "as.Date"
    )
  )

})

test_that(".listPermittedTidyverseFunctionsDS allows functions to be specified via options", {

  options("tidyverse.permitted.functions" = c("function1", "function2"))
  expect_equal(
    listPermittedTidyverseFunctionsDS(),
    c("function1", "function2")
  )

})

test_that(".check_function_names works with functions which contain a period", {

  options("tidyverse.permitted.functions" = "as.factor")
  expect_silent(.check_function_names("as.factor(cyl)"))

})

# Reset permitted functions to default for .build_eval_env tests
options("tidyverse.permitted.functions" = NULL)

test_that(".build_eval_env contains permitted dplyr functions", {
  eval_env <- .build_eval_env(globalenv())
  expect_true(exists("desc", envir = eval_env, inherits = FALSE))
  expect_true(exists("lag", envir = eval_env, inherits = FALSE))
  expect_true(exists("if_else", envir = eval_env, inherits = FALSE))
  expect_true(exists("case_when", envir = eval_env, inherits = FALSE))
  expect_true(exists("mutate", envir = eval_env, inherits = FALSE))
})

test_that(".build_eval_env does not contain base R functions", {
  eval_env <- .build_eval_env(globalenv())
  expect_false(exists("mean", envir = eval_env, inherits = FALSE))
  expect_false(exists("sqrt", envir = eval_env, inherits = FALSE))
  expect_false(exists("is.na", envir = eval_env, inherits = FALSE))
})

test_that(".build_eval_env does not contain functions from other packages", {
  eval_env <- .build_eval_env(globalenv())
  expect_false(exists("str_detect", envir = eval_env, inherits = FALSE))
  expect_false(exists("ggplot", envir = eval_env, inherits = FALSE))
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
login_data <- .prepare_dslite("filterDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

good_filter_arg <- "mpg > 20 & gear > 2"
mtcars_group <- mtcars %>% group_by(cyl)

test_that("filterDS correctly filters where data and columns exist", {
  good_filter_cally <- .make_tidyverse_call("mtcars", "filter", good_filter_arg)
  expect_equal(
    dim(eval(good_filter_cally))[[1]],
    14
  )
})

test_that("filterDS works with .by argument", {
  by_cally <- .make_tidyverse_call("mtcars", "filter", "mpg > median(mpg)", ".by = 'cyl'")
  no_by_cally <- .make_tidyverse_call("mtcars", "filter", "mpg > median(mpg)")

  filter_with_by <- eval(by_cally)
  filter_without_by <- eval(no_by_cally)

  expect_false(
    identical(
      rownames(filter_with_by),
      rownames(filter_without_by)
    )
  )

  expect_equal(
    dim(filter_with_by),
    c(14, 11)
  )
})

test_that("filterDS works with .preserve argument", {
  preserve_cally_true <- .make_tidyverse_call("mtcars_group", "filter", good_filter_arg, ".preserve = TRUE")
  expect_equal(
    group_keys(eval(preserve_cally_true)),
    tibble(cyl = c(4, 6, 8))
  )

  preserve_cally_false <- .make_tidyverse_call("mtcars_group", "filter", good_filter_arg, ".preserve = FALSE")
  expect_equal(
    group_keys(eval(preserve_cally_false)),
    tibble(cyl = c(4, 6))
  )
})

test_that("filterDS fails when data doesn't exist", {
  no_data <- .make_tidyverse_call("doesntexist", "filter", good_filter_arg)
  expect_error(
    eval(no_data),
    "`filterDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'doesntexist'\\s+not\\s+found"
  )
})

test_that("filterDS passes with valid combinations of extra arguments", {
  just_by <- .make_tidyverse_call("mtcars", "filter", good_filter_arg, ".by = 'carb'")
  just_preserve <- .make_tidyverse_call("mtcars_group", "filter", good_filter_arg, ".preserve = T")
  by_preserve_f <- .make_tidyverse_call("mtcars", "filter", good_filter_arg, c(".by = 'carb', .preserve = FALSE"))

  expect_silent(eval(just_by))
  expect_silent(eval(just_preserve))
  expect_silent(eval(by_preserve_f))
})

test_that("filterDS fails if supplied with .by and .preserve = T", {
  by_preserve_t <- .make_tidyverse_call("mtcars", "filter", good_filter_arg, c(".by = 'carb', .preserve = T"))
  expect_error(
    eval(by_preserve_t),
    "Can't supply both `.by` and `.preserve`"
  )
})

test_that("filterDS fails with bad argument", {
  bad_filter_arg <- "test_1 = mpg, mutate/asd"
  bad_call <- .make_tidyverse_call("mtcars", "filter", bad_filter_arg)
  expect_error(
    eval(bad_call),
    "We detected a named input\\."
  )
})

test_that("filterDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("filterDS", "carb$SPACE$$EQU$$EQU$$SPACE$4", "mtcars", NULL, FALSE)

  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.dim("test", datasources = conns)[[1]],
    c(10, 11)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
login_data <- .prepare_dslite("ifElseDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

if_else_num_arg <- "mtcars$mpg > 20"

test_that("ifElseDS passes and numeric condition and categorial output", {
  other_args <- "true = \"high\", false = \"low\", missing = NULL, ptype = NULL, size = NULL"
  good_if_else_cally <- .make_tidyverse_call("mtcars", "if_else", if_else_num_arg, other_args, inc_data = F)
  result <- eval(good_if_else_cally)

  expect_equal(
    result,
    c(
      "high", "high", "high", "high", "low", "low", "low", "high", "high", "low", "low", "low",
      "low", "low", "low", "low", "low", "high", "high", "high", "high", "low", "low", "low",
      "low", "high", "high", "high", "low", "low", "low", "high"
    )
  )
})

test_that("ifElseDS passes and numeric condition and numeric output", {
  other_args <- "true = 50, false = 100, missing = NULL, ptype = NULL, size = NULL"
  good_if_else_cally <- .make_tidyverse_call("mtcars", "if_else", if_else_num_arg, other_args, inc_data = F)
  result <- eval(good_if_else_cally)

  expect_equal(
    result,
    c(
      50, 50, 50, 50, 100, 100, 100, 50, 50, 100, 100, 100,
      100, 100, 100, 100, 100, 50, 50, 50, 50, 100, 100, 100,
      100, 50, 50, 50, 100, 100, 100, 50
    )
  )
})

test_that("ifElseDS passes with categorical condition", {
  mtcars_cat <- mtcars %>% mutate(mpg_cat = if_else(mpg > 20, "high", "low"))
  if_else_cat_arg <- "mtcars_cat$mpg_cat == \"high\""
  other_args <- "true = 1000, false = 10, missing = NULL, ptype = NULL, size = NULL"
  good_mutate_cally <- .make_tidyverse_call("mtcars_cat", "if_else", if_else_cat_arg, other_args, inc_data = F)
  result <- eval(good_mutate_cally)

  expect_equal(
    result,
    c(
      1000, 1000, 1000, 1000, 10, 10, 10, 1000, 1000, 10, 10, 10, 10, 10, 10, 10, 10,
      1000, 1000, 1000, 1000, 10, 10, 10, 10, 1000, 1000, 1000, 10, 10, 10, 1000
    )
  )
})

test_that("ifElseDS passes when `missing` argument used", {
  other_args <- "true = \"high\", false = \"low\", missing = \"val_missing\", ptype = NULL, size = NULL"
  mtcars <- mtcars %>% mutate(mpg = NA)
  good_if_else_cally <- .make_tidyverse_call("mtcars", "if_else", if_else_num_arg, other_args, inc_data = F)
  result <- eval(good_if_else_cally)

  expect_equal(
    result,
    rep("val_missing", 32)
  )
})

test_that("ifElseDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("ifElseDS", "mtcars$mpg$SPACE$$GT$$SPACE$20", "high", "low", NULL, NULL, NULL)
  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "character"
  )

  expect_equal(
    as.numeric(ds.table("test", datasources = conns)$output.list$TABLES.COMBINED_all.sources_counts),
    c(42, 54, 0)
  )
})

test_that("ifElseDS passes when `ptype` argument used", {
  data(mtcars)
  other_args <- "true = \"50\", false = \"500\", missing = NULL, ptype = \"character\", size = NULL"
  good_if_else_cally <- .make_tidyverse_call("mtcars_mis", "if_else", if_else_num_arg, other_args, inc_data = F)
  result <- eval(good_if_else_cally)

  expect_equal(
    result,
    c(
      "50", "50", "50", "50", "500", "500", "500", "50", "50",
      "500", "500", "500", "500", "500", "500", "500", "500", "50",
      "50", "50", "50", "500", "500", "500", "500", "50", "50",
      "50", "500", "500", "500", "50"
    )
  )
})

test_that("ifElseDS passes when `size` argument used", {
  other_args <- "true = \"high\", false = \"low\", missing = NULL, ptype = NULL, size = 32"
  good_if_else_cally <- .make_tidyverse_call("mtcars_mis", "if_else", if_else_num_arg, other_args, inc_data = F)
  result <- eval(good_if_else_cally)

  expect_equal(
    result,
    c(
      "high", "high", "high", "high", "low", "low", "low", "high", "high", "low", "low", "low",
      "low", "low", "low", "low", "low", "high", "high", "high", "high", "low", "low", "low",
      "low", "high", "high", "high", "low", "low", "low", "high"
    )
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
login_data <- .prepare_dslite("caseWhenDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

test_that("caseWhenDS passes and numeric condition and categorical output", {
  case_when_arg <-
    'mtcars$mpg < 10 ~ "low",
  mtcars$mpg >= 10 & mtcars$mpg < 20 ~ "medium",
  mtcars$mpg >= 20 ~ "high"'

  other_args <- ".default = NULL, .ptype = NULL, .size = NULL"
  case_when_cally <- .make_tidyverse_call("mtcars", "case_when", case_when_arg, other_args, inc_data = F)
  result <- eval(case_when_cally)

  expect_equal(
    result,
    c(
      "high", "high", "high", "high", "medium", "medium", "medium", "high",
      "high", "medium", "medium", "medium", "medium", "medium", "medium",
      "medium", "medium", "high", "high", "high", "high", "medium", "medium",
      "medium", "medium", "high", "high", "high", "medium", "medium", "medium", "high"
    )
  )
})

test_that("caseWhenDS passes and numeric condition and numeric output", {
  case_when_arg <-
    "mtcars$mpg < 10 ~ 10,
  mtcars$mpg >= 10 & mtcars$mpg < 20 ~ 20,
  mtcars$mpg >= 20 ~ 30"

  other_args <- ".default = NULL, .ptype = NULL, .size = NULL"
  case_when_cally <- .make_tidyverse_call("mtcars", "case_when", case_when_arg, other_args, inc_data = F)
  result <- eval(case_when_cally)

  expect_equal(
    result,
    c(
      30, 30, 30, 30, 20, 20, 20, 30, 30, 20, 20, 20, 20, 20, 20, 20, 20, 30, 30, 30, 30, 20, 20,
      20, 20, 30, 30, 30, 20, 20, 20, 30
    )
  )
})

test_that("caseWhenDS passes with categorical condition and categorical output", {
  case_when_arg <-
    'mtcars$gear == 1 ~ "low",
     mtcars$gear == 2 ~ "medium",
     mtcars$gear == 3 ~ "high",
     mtcars$gear == 4 ~ "very_high"'

  other_args <- ".default = NULL, .ptype = NULL, .size = NULL"
  case_when_cally <- .make_tidyverse_call("mtcars", "case_when", case_when_arg, other_args, inc_data = F)
  result <- eval(case_when_cally)

  expect_equal(
    result,
    c(
      "very_high", "very_high", "very_high", "high", "high", "high", "high", "very_high", "very_high", "very_high",
      "very_high", "high", "high", "high", "high", "high", "high", "very_high", "very_high", "very_high",
      "high", "high", "high", "high", "high", "very_high", NA, NA, NA, NA,
      NA, "very_high"
    )
  )
})

test_that("caseWhenDS passes with .default argument", {
  case_when_arg <-
    'mtcars$gear == 1 ~ "low",
     mtcars$gear == 2 ~ "medium",
     mtcars$gear == 3 ~ "high",
     mtcars$gear == 4 ~ "very_high"'

  other_args <- ".default = 'something_missing', .ptype = NULL, .size = NULL"
  case_when_cally <- .make_tidyverse_call("mtcars", "case_when", case_when_arg, other_args, inc_data = F)
  result <- eval(case_when_cally)

  expect_equal(
    result,
    c(
      "very_high", "very_high", "very_high", "high", "high", "high",
      "high", "very_high", "very_high", "very_high", "very_high", "high",
      "high", "high", "high", "high", "high", "very_high",
      "very_high", "very_high", "high", "high", "high",
      "high", "high", "very_high", "something_missing",
      "something_missing", "something_missing", "something_missing",
      "something_missing", "very_high"
    )
  )
})

test_that("caseWhenDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("caseWhenDS", "mtcars$mpg$SPACE$$LT$$SPACE$20$SPACE$$TILDE$$SPACE$$QUOTE$low$QUOTE$$COMMA$$SPACE$mtcars$mpg$SPACE$$GT$$EQU$$SPACE$20$SPACE$$AND$$SPACE$mtcars$mpg$SPACE$$LT$$SPACE$$LINE$$SPACE$$SPACE$$SPACE$$SPACE$30$SPACE$$TILDE$$SPACE$$QUOTE$medium$QUOTE$$COMMA$$SPACE$mtcars$mpg$SPACE$$GT$$EQU$$SPACE$30$SPACE$$TILDE$$SPACE$$QUOTE$high$QUOTE$", NULL, NULL, NULL)
  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "character"
  )

  expect_equal(
    as.numeric(ds.table("test", datasources = conns)$output.list$TABLES.COMBINED_all.sources_counts),
    c(12, 54, 30, 0)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
login_data <- .prepare_dslite("arrangeDS", NULL, list(mtcars = mtcars, mtcars_group = mtcars_group))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")
datashield.assign.table(conns, "mtcars_group", "mtcars_group")

good_arrange_arg <- "mpg, cyl"

test_that("arrangeDS correctly arranges data", {
  good_arrange_cally <- .make_tidyverse_call("mtcars", "arrange", good_arrange_arg)

  sorted_df <- eval(good_arrange_cally)

  expect_equal(
    class(sorted_df),
    "data.frame"
  )

  expect_equal(
    sorted_df$mpg,
    c(
      10.4, 10.4, 13.3, 14.3, 14.7, 15.0, 15.2, 15.2, 15.5, 15.8, 16.4, 17.3, 17.8, 18.1, 18.7,
      19.2, 19.2, 19.7, 21.0, 21.0, 21.4, 21.4, 21.5, 22.8, 22.8, 24.4, 26.0, 27.3, 30.4, 30.4,
      32.4, 33.9
    )
  )

  expect_equal(
    sorted_df$cyl,
    c(
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 6, 6, 8, 6, 8, 6, 6, 6, 4, 6, 4, 4, 4, 4, 4, 4, 4, 4, 4,
      4
    )
  )
})


test_that("arrangeDS works with .by_group argument", {
  by_cally <- .make_tidyverse_call("mtcars_group", "arrange", "drat", ".by_group = TRUE")
  no_by_cally <- .make_tidyverse_call("mtcars_group", "arrange", "drat", ".by_group = FALSE")

  arrange_by_t <- eval(by_cally)
  arrange_by_f <- eval(no_by_cally)

  expect_equal(
    class(arrange_by_t),
    c("grouped_df", "tbl_df", "tbl", "data.frame")
  )

  expect_equal(
    class(arrange_by_f),
    c("grouped_df", "tbl_df", "tbl", "data.frame")
  )

  expect_equal(
    arrange_by_t$drat,
    c(
      3.69, 3.70, 3.77, 3.85, 3.92, 4.08, 4.08, 4.11, 4.22, 4.43, 4.93, 2.76, 3.08, 3.62, 3.90,
      3.90, 3.92, 3.92, 2.76, 2.93, 3.00, 3.07, 3.07, 3.07, 3.08, 3.15, 3.15, 3.21, 3.23, 3.54,
      3.73, 4.22
    )
  )

  expect_equal(
    arrange_by_f$drat,
    c(
      2.76, 2.76, 2.93, 3.00, 3.07, 3.07, 3.07, 3.08, 3.08, 3.15, 3.15, 3.21, 3.23, 3.54, 3.62,
      3.69, 3.70, 3.73, 3.77, 3.85, 3.90, 3.90, 3.92, 3.92, 3.92, 4.08, 4.08, 4.11, 4.22, 4.22,
      4.43, 4.93
    )
  )
})

test_that("arrangeDS works with desc option", {
  desc_arrange_arg <- "list(desc(mpg))"
  desc_arrange_cally <- .make_tidyverse_call("mtcars", "arrange", desc_arrange_arg)

  desc_df <- eval(desc_arrange_cally)

  expect_equal(
    class(desc_df),
    "data.frame"
  )

  expect_equal(
    desc_df$mpg,
    c(
      21.0, 21.0, 22.8, 21.4, 18.7, 18.1, 14.3, 24.4, 22.8, 19.2, 17.8,
      16.4, 17.3, 15.2, 10.4, 10.4, 14.7, 32.4, 30.4, 33.9, 21.5,
      15.5, 15.2, 13.3, 19.2, 27.3, 26.0, 30.4, 15.8, 19.7, 15.0, 21.4
    )
  )
})

test_that("arrangeDS fails when data doesn't exist", {
  no_data <- .make_tidyverse_call("doesnt_exist", "arrange", good_arrange_arg)
  expect_error(
    eval(no_data),
    "object 'doesnt_exist' not found"
  )
})

test_that("arrangeDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("arrangeDS", "drat", "mtcars", NULL)
  datashield.assign(conns, "sorted_df", cally)

  expect_equal(
    ds.class("sorted_df", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.dim("sorted_df", datasources = conns)[[1]],
    c(32, 11)
  )
})

test_that("arrangeDS works with desc option when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("arrangeDS", "desc$LB$drat$RB$", "mtcars", NULL)
  datashield.assign(conns, "sorted_df", cally)

  expect_equal(
    ds.class("sorted_df", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.dim("sorted_df", datasources = conns)[[1]],
    c(32, 11)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
login_data <- .prepare_dslite("sliceDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

mtcars_group <- mtcars %>% group_by(carb)

good_slice_arg <- "1:5"

test_that("sliceDS correctly filters where data and columns exist", {
  good_slice_cally <- .make_tidyverse_call("mtcars", "slice", good_slice_arg)
  expect_equal(
    dim(eval(good_slice_cally)),
    c(5, 11)
  )
})

test_that("sliceDS works with .by argument", {
  by_cally <- .make_tidyverse_call("mtcars", "slice", "1:5", ".by = 'cyl'")
  no_by_cally <- .make_tidyverse_call("mtcars", "slice", "1:5")

  slice_with_by <- eval(by_cally)
  slice_without_by <- eval(no_by_cally)

  expect_false(
    identical(
      rownames(slice_with_by),
      rownames(slice_without_by)
    )
  )

  expect_equal(
    dim(slice_with_by),
    c(15, 11)
  )

  expect_equal(
    dim(slice_without_by),
    c(5, 11)
  )
})

test_that("sliceDS works with .preserve argument", {
  preserve_cally_true <- .make_tidyverse_call("mtcars_group", "slice", "9:12", ".preserve = TRUE")
  expect_equal(
    group_keys(eval(preserve_cally_true)),
    tibble(carb = c(1, 2, 3, 4, 6, 8))
  )

  preserve_cally_false <- .make_tidyverse_call("mtcars_group", "slice", "9:12", ".preserve = FALSE")
  expect_equal(
    group_keys(eval(preserve_cally_false)),
    tibble(carb = c(2, 4))
  )
})

test_that("sliceDS fails when data doesn't exist", {
  no_data <- .make_tidyverse_call("doesntexist", "slice", good_slice_arg)
  expect_error(
    eval(no_data),
    "object 'doesntexist' not found"
  )
})

test_that("sliceDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("sliceDS", "1:5", "mtcars", NULL, FALSE)

  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.dim("test", datasources = conns)[[1]],
    c(5, 11)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

options(datashield.errors.print = TRUE)
data("mtcars")
login_data <- .prepare_dslite("distinctDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

empty_distinct_arg <- character(0)

test_that("distinctDS correctly works where no `expr` argument is provided", {
  empty_distinct_cally <- .make_tidyverse_call("mtcars", "distinct", empty_distinct_arg)

  out_empty <- eval(empty_distinct_cally)
  expect_equal(
    class(out_empty),
    "data.frame"
  )

  expect_equal(
    dim(out_empty),
    c(32, 11)
  )
})

test_that("distinctDS correctly works where variables specified in `expr` argument", {
  expr_distinct_cally <- .make_tidyverse_call("mtcars", "distinct", "cyl, drat")

  out_expr <- eval(expr_distinct_cally)

  expect_equal(
    class(out_expr),
    "data.frame"
  )

  expect_equal(
    dim(out_expr),
    c(26, 2)
  )
})

test_that("distinctDS correctly works with `.keep_all` argument", {
  keep_all_y_cally <- .make_tidyverse_call("mtcars", "distinct", "cyl, drat", ".keep_all = TRUE")
  keep_all_y_out <- eval(keep_all_y_cally)

  expect_equal(
    class(keep_all_y_out),
    "data.frame"
  )

  expect_equal(
    dim(keep_all_y_out),
    c(26, 11)
  )

  expect_equal(
    colnames(keep_all_y_out),
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")
  )

  keep_all_n_cally <- .make_tidyverse_call("mtcars", "distinct", "cyl, drat", ".keep_all = FALSE")
  keep_all_n_out <- eval(keep_all_n_cally)

  expect_equal(
    class(keep_all_n_out),
    "data.frame"
  )

  expect_equal(
    dim(keep_all_n_out),
    c(26, 2)
  )

  expect_equal(
    colnames(keep_all_n_out),
    c("cyl", "drat")
  )
})

test_that("distinctDS fails when data doesn't exist", {
  no_data <- .make_tidyverse_call("doesntexist", "distinct", empty_distinct_arg)
  expect_error(
    eval(no_data),
    "object 'doesntexist' not found"
  )
})

test_that("distinctDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("distinctDS", "mpg$COMMA$$SPACE$cyl", "mtcars", FALSE)
  datashield.assign(conns, "test_distinct", cally)

  expect_equal(
    ds.class("test_distinct", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.dim("test_distinct", datasources = conns)[[1]],
    c(27, 2)
  )
})

test_that("distinctDS fails if subset is too small", {
  cally <- call("distinctDS", "mpg$COMMA$$SPACE$cyl$COMMA$$SPACE$drat", "mtcars", FALSE)
  expect_error(
    datashield.assign(conns, "test_distinct", cally)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_dup_names <- cbind(mtcars, tibble(cyl = 2))
test_matrix <- matrix(data = 1:20, ncol = 4)

login_data <- .prepare_dslite(
  assign_method = "asTibbleDS",
  tables = list(
    mtcars = mtcars,
    mtcars_dup_names = mtcars_dup_names,
    test_matrix = test_matrix
  )
)

conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")
datashield.assign.table(conns, "mtcars_dup_names", "mtcars_dup_names")
datashield.assign.table(conns, "test_matrix", "test_matrix")

tibble_class <- c("tbl_df", "tbl", "data.frame")

test_that("asTibbleDS correctly converts a data frame to a tibble", {
  good_tibble_cally <- .make_tidyverse_call("mtcars", "as_tibble", tidy_expr = NULL)

  now_a_tibble <- eval(good_tibble_cally)

  expect_equal(
    class(now_a_tibble),
    tibble_class
  )

  expect_equal(
    dim(now_a_tibble),
    c(32, 11)
  )
})

test_that("asTibbleDS correctly converts a matrix to a tibble", {
  good_matrix_cally <- .make_tidyverse_call("test_matrix", "as_tibble", tidy_expr = NULL, ".name_repair = 'minimal'")

  matrix_to_tibble <- eval(good_matrix_cally)

  expect_equal(
    class(matrix_to_tibble),
    tibble_class
  )

  expect_equal(
    dim(matrix_to_tibble),
    c(5, 4)
  )
})

test_that("asTibbleDS fails when data doesn't exist", {
  no_data <- .make_tidyverse_call("doesnt_exist", "as_tibble", tidy_expr = NULL)
  expect_error(
    eval(no_data),
    "object 'doesnt_exist' not found"
  )
})


test_that("asTibbleDS works with the name_repair argument", {
  repair_min_call <- .make_tidyverse_call("mtcars_dup_names", "as_tibble", tidy_expr = NULL, '.name_repair = "minimal"')
  tib_min_repair <- eval(repair_min_call)

  expect_equal(
    class(tib_min_repair),
    tibble_class
  )

  expect_equal(
    colnames(tib_min_repair),
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "cyl")
  )

  repair_unique_call <- .make_tidyverse_call("mtcars_dup_names", "as_tibble", tidy_expr = NULL, '.name_repair = "unique"')

  expect_message(
    tib_unique_repair <- eval(repair_unique_call)
  )

  expect_equal(
    class(tib_unique_repair),
    tibble_class
  )

  expect_equal(
    colnames(tib_unique_repair),
    c("mpg", "cyl...2", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "cyl...12")
  )

  repair_check_call <- .make_tidyverse_call("mtcars_dup_names", "as_tibble", tidy_expr = NULL, '.name_repair = "check_unique"')

  expect_error(
    tib_unique_repair <- eval(repair_check_call)
  )

  repair_uni_call <- .make_tidyverse_call("mtcars_dup_names", "as_tibble", tidy_expr = NULL, '.name_repair = "universal"')

  expect_message(
    tib_uni_repair <- eval(repair_uni_call)
  )

  expect_equal(
    class(tib_uni_repair),
    tibble_class
  )

  expect_equal(
    colnames(tib_uni_repair),
    c("mpg", "cyl...2", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "cyl...12")
  )
})

test_that("asTibbleDS works with the rownames argument", {
  row_null_call <- .make_tidyverse_call("mtcars", "as_tibble", tidy_expr = NULL, "rownames = NULL")
  row_null_tibble <- eval(row_null_call)

  expect_equal(
    class(row_null_tibble),
    tibble_class
  )

  expect_equal(
    rownames(row_null_tibble),
    as.character(rep(1:32))
  )

  row_na_call <- .make_tidyverse_call("mtcars", "as_tibble", tidy_expr = NULL, "rownames = NA")
  row_na_tibble <- eval(row_na_call)

  expect_equal(
    class(row_na_tibble),
    tibble_class
  )

  car_names <- c(
    "Mazda RX4", "Mazda RX4 Wag", "Datsun 710", "Hornet 4 Drive",
    "Hornet Sportabout", "Valiant", "Duster 360", "Merc 240D", "Merc 230",
    "Merc 280", "Merc 280C", "Merc 450SE", "Merc 450SL", "Merc 450SLC",
    "Cadillac Fleetwood", "Lincoln Continental", "Chrysler Imperial",
    "Fiat 128", "Honda Civic", "Toyota Corolla", "Toyota Corona",
    "Dodge Challenger", "AMC Javelin", "Camaro Z28", "Pontiac Firebird",
    "Fiat X1-9", "Porsche 914-2", "Lotus Europa", "Ford Pantera L",
    "Ferrari Dino", "Maserati Bora", "Volvo 142E"
  )

  expect_equal(
    rownames(row_na_tibble),
    car_names
  )

  row_col_call <- .make_tidyverse_call("mtcars", "as_tibble", tidy_expr = NULL, 'rownames = "col_with_names"')
  tib_with_col <- eval(row_col_call)

  expect_equal(
    class(tib_with_col),
    tibble_class
  )

  expect_equal(
    colnames(tib_with_col),
    c(
      "col_with_names", "mpg", "cyl", "disp", "hp", "drat",
      "wt", "qsec", "vs", "am", "gear", "carb"
    )
  )

  expect_equal(
    tib_with_col$col_with_names,
    car_names
  )
})

test_that("asTibbleDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("asTibbleDS", NULL, "mtcars", NULL, "minimal", NULL)
  datashield.assign(conns, "new_tibble", cally)

  expect_equal(
    ds.class("new_tibble", datasources = conns)[[1]],
    tibble_class
  )

  expect_equal(
    ds.colnames("new_tibble", datasources = conns)[[1]],
    c(
      "mpg", "cyl", "disp", "hp", "drat",
      "wt", "qsec", "vs", "am", "gear", "carb"
    )
  )

  expect_equal(
    ds.dim("new_tibble", datasources = conns)[[1]],
    c(32, 11)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>%
  group_by(cyl) %>%
  mutate(drop_test = factor("a", levels = c("a", "b")))
login_data <- .prepare_dslite("groupByDS", NULL, list(mtcars = mtcars, mtcars_group = mtcars_group))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")
datashield.assign.table(conns, "mtcars_group", "mtcars_group")

test_that("groupByDS correctly groups data where data and columns exist", {
  good_group_expr <- "cyl"
  group_call <- .make_tidyverse_call("mtcars", "group_by", good_group_expr)

  expect_equal(
    group_keys(eval(group_call)),
    tibble(cyl = c(4, 6, 8))
  )

  mult_group_expr <- "cyl, mpg"
  mult_call <- .make_tidyverse_call("mtcars", "group_by", mult_group_expr)

  expect_equal(
    group_keys(eval(mult_call)),
    tibble(
      cyl = c(rep(4, 9), rep(6, 6), rep(8, 12)),
      mpg = c(
        21.4, 21.5, 22.8, 24.4, 26, 27.3, 30.4, 32.4, 33.9,
        17.8, 18.1, 19.2, 19.7, 21, 21.4,
        10.4, 13.3, 14.3, 14.7, 15, 15.2, 15.5, 15.8, 16.4, 17.3, 18.7, 19.2
      )
    )
  )
})

test_that("groupByDS works with .add argument", {
  add_true_call <- .make_tidyverse_call("mtcars_group", "group_by", "gear", ".add = TRUE")
  grouped_add_true <- eval(add_true_call)
  expect_equal(
    group_keys(grouped_add_true),
    tibble(cyl = c(4, 4, 4, 6, 6, 6, 8, 8), gear = c(3, 4, 5, 3, 4, 5, 3, 5))
  )

  add_false_call <- .make_tidyverse_call("mtcars_group", "group_by", "cyl", ".add = FALSE")
  grouped_add_false <- eval(add_false_call)
  expect_equal(
    group_keys(grouped_add_false),
    tibble(cyl = c(4, 6, 8))
  )
})

test_that("groupByDS works with .drop argument", {
  drop_true_call <- .make_tidyverse_call("mtcars_group", "group_by", "drop_test", ".drop = TRUE")
  grouped_drop_true <- eval(drop_true_call)
  expect_equal(
    group_keys(grouped_drop_true)$drop_test,
    factor("a", levels = c("a", "b"))
  )

  drop_false_call <- .make_tidyverse_call("mtcars_group", "group_by", "drop_test", ".drop = FALSE")
  grouped_drop_false <- eval(drop_false_call)
  expect_equal(
    group_keys(grouped_drop_false)$drop_test,
    factor(c("a", "b"), levels = c("a", "b"))
  )
})

test_that("groupByDS fails when data doesn't exist", {
  no_data_call <- .make_tidyverse_call("doesntexist", "group_by", "cyl")
  expect_error(
    eval(no_data_call),
    "object 'doesntexist' not found"
  )
})

test_that("groupByDS fails with bad argument", {
  bad_group_expr <- "wrong_column"
  bad_call <- .make_tidyverse_call("mtcars", "group_by", bad_group_expr)
  expect_error(
    eval(bad_call),
    "Must group by variables found in `.data`"
  )
})

test_that("groupByDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  call_direct <- call("groupByDS", "cyl", "mtcars", FALSE, TRUE)
  datashield.assign(conns, "test_group", call_direct)

  expect_equal(
    ds.class("test_group", datasources = conns)[[1]],
    c("grouped_df", "tbl_df", "tbl", "data.frame")
  )
})


add_true_call <- .make_tidyverse_call("mtcars_group", "group_by", "gear", ".add = TRUE")
grouped_add_true <- eval(add_true_call)


test_that("ungroupDS correctly ungroups data", {
  ungroup_call <- .make_tidyverse_call("mtcars_group", "ungroup", tidy_expr = NULL, other_args = NULL)
  ungrouped_data <- eval(ungroup_call)

  expect_equal(
    class(ungrouped_data),
    c("tbl_df", "tbl", "data.frame")
  )
})

test_that("ungroupDS works with already ungrouped data", {
  ungroup_call <- .make_tidyverse_call("mtcars", "ungroup", tidy_expr = NULL, other_args = NULL)
  ungrouped_data <- eval(ungroup_call)

  expect_equal(
    class(ungrouped_data),
    "data.frame"
  )
})

test_that("ungroupDS fails when data doesn't exist", {
  no_data_call <- .make_tidyverse_call("doesntexist", "ungroup", tidy_expr = NULL, other_args = NULL)
  expect_error(
    eval(no_data_call),
    "object 'doesntexist' not found"
  )
})

data("mtcars")
mtcars_group <- mtcars %>%
  group_by(cyl) %>%
  mutate(drop_test = factor("a", levels = c("a", "b")))
login_data <- .prepare_dslite("ungroupDS", NULL, list(mtcars_group = mtcars_group))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars_group", "mtcars_group")

test_that("ungroupDS works correctly when called directly", {
  skip_if_not_installed("dsBaseClient")
  ungroup_call <- call("ungroupDS", NULL, "mtcars_group")
  datashield.assign(conns, "ungrouped_data", ungroup_call)

  expect_equal(
    ds.class("ungrouped_data", datasources = conns)[[1]],
    c("tbl_df", "tbl", "data.frame")
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
login_data <- .prepare_dslite("selectDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

good_select_arg <- "mpg, cyl, starts_with('g'), ends_with('b')"

test_that("selectDS passes where data and column exist", {
  good_select_cally <- .make_tidyverse_call("mtcars", "select", good_select_arg)
  expected <- c("mpg", "cyl", "gear", "carb")
  expect_equal(
    colnames(eval(good_select_cally)),
    expected
  )
})

call_no_data <- .make_tidyverse_call("data_not_there", "select", good_select_arg)

test_that("selectDS fails with correct message when data doesn't exist", {
  select_cally_no_data <- .make_tidyverse_call("doesntexist", "select", good_select_arg)
  expect_error(
    eval(select_cally_no_data),
    "object 'doesntexist' not found"
  )
})

test_that(".execute_tidyverse_function fails with correct message when unrecognised function passed", {
  select_random_arg <- "filter('mpg'), mpg, cyl, starts_with('g'), ends_with('b')"
  random_select_cally <- .make_tidyverse_call("mtcars", "select", select_random_arg)
  expect_error(
    eval(random_select_cally),
    "`selectDS`\\s+returned\\s+the\\s+following\\s+error:|no\\s+applicable\\s+method\\s+for\\s+'filter'"
  )
})

test_that("select passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call(
    "selectDS", "mpg$COMMA$$SPACE$starts_with$LB$$QUOTE$m$QUOTE$$RB$$COMMA$$SPACE$ends_with$LB$$QUOTE$t$QUOTE$$RB$",
    "mtcars"
  )
  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.colnames("test", datasources = conns)[[1]],
    c("mpg", "drat", "wt")
  )
})

test_that("select passes strings that included 'c'", {
  good_arg_1 <- "any_of(c(mpg, cyl, drat))"
  expect_silent(
    .check_tidy_disclosure("mtcars", good_arg_1)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)
require(purrr)

data("mtcars")
login_data <- .prepare_dslite("bindRowsDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

test_that("bindRowsDS passes", {
  bind_rows_arg <- "mtcars, mtcars"
  other_args <- ".id = NULL"
  bind_rows_cally <- .make_tidyverse_call(NULL, "bind_rows", bind_rows_arg, other_args, inc_data = F)
  result <- eval(bind_rows_cally)

  expect_equal(
    class(result),
    "data.frame"
  )

  expect_equal(
    colnames(result),
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")
  )

  expect_equal(
    dim(result),
    c(64, 11)
  )

  expect_equal(
    result %>% map_dbl(mean) %>% round(2),
    c(mpg = 20.09, cyl = 6.19, disp = 230.72, hp = 146.69, drat = 3.60, wt = 3.22, qsec = 17.85, vs = 0.44, am = 0.41, gear = 3.69, carb = 2.81)
  )
})

test_that("bindRowsDS passes with .id argument", {
  bind_rows_arg <- "mtcars, mtcars"
  other_args <- ".id = 'test_col_name'"
  bind_rows_cally <- .make_tidyverse_call(NULL, "bind_rows", bind_rows_arg, other_args, inc_data = F)
  result <- eval(bind_rows_cally)

  expect_equal(
    colnames(result),
    c("test_col_name", "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")
  )
})

test_that("bindRowsDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("bindRowsDS", "mtcars$COMMA$$SPACE$mtcars", NULL)
  datashield.assign(conns, "test", cally)


  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.colnames("test", datasources = conns)[[1]],
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")
  )

  expect_equal(
    ds.dim("test", datasources = conns)[[1]],
    c(64, 11)
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
date_df <- data.frame(date_str = c("2024-01-15", "2024-06-30", "2024-12-25"), stringsAsFactors = FALSE)
login_data <- .prepare_dslite("mutateDS", NULL, list(mtcars = mtcars, date_df = date_df))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")
datashield.assign.table(conns, "date_df", "date_df")

options(
  datashield.privacyControlLevel = "banana",
  nfilter.tab = 3,
  nfilter.subset = 3,
  nfilter.glm = 0.33,
  nfilter.string = 80,
  nfilter.stringShort = 20,
  nfilter.kNN = 3,
  nfilter.levels.density = 0.33,
  nfilter.levels.max = 40,
  nfilter.noise = 0.25,
  nfilter.privacy.old = 5
)

good_mutate_arg <- "mpg_trans = cyl*1000, new_var = (hp-drat)/qsec"
# good_mutate_arg_enc <- .encode_tidy_eval(good_mutate_arg, .get_encode_dictionary())

test_that("mutateDS passes where data and column exist", {
  good_mutate_cally <- .make_tidyverse_call("mtcars", "mutate", good_mutate_arg, list(".keep = \"all\", .before = NULL, .after = NULL"))
  result <- eval(good_mutate_cally)

  expect_equal(
    mean(result$mpg_trans),
    6187.5
  )

  expect_equal(
    colnames(result),
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "mpg_trans", "new_var")
  )
})

test_that("mutateDS passes with different .keep argument", {
  good_mutate_cally <- .make_tidyverse_call("mtcars", "mutate", good_mutate_arg, list(".keep = \"none\", .before = NULL, .after = NULL"))
  result <- eval(good_mutate_cally)

  expect_equal(
    colnames(result),
    c("mpg_trans", "new_var")
  )
})

test_that("mutateDS passes with different .before argument", {
  good_mutate_cally <- .make_tidyverse_call("mtcars", "mutate", good_mutate_arg, list(".keep = \"all\", .before = \"disp\", .after = NULL"))
  result <- eval(good_mutate_cally)

  expect_equal(
    colnames(result),
    c("mpg", "cyl", "mpg_trans", "new_var", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb")
  )
})

test_that("mutateDS passes with different .afterb argument", {
  good_mutate_cally <- .make_tidyverse_call("mtcars", "mutate", good_mutate_arg, list(".keep = \"all\", .before = NULL, .after = \"qsec\""))
  result <- eval(good_mutate_cally)

  expect_equal(
    colnames(result),
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "mpg_trans", "new_var", "vs", "am", "gear", "carb")
  )
})

test_that("mutateDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call(
    "mutateDS", "new_var$SPACE$$EQU$$SPACE$mpg$SPACE$$MULT$$SPACE$1000$COMMA$$SPACE$new_var2$SPACE$$EQU$$SPACE$hp$SPACE$$SUB$$SPACE$drat",
    "mtcars", "all", NULL, NULL
  )

  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.colnames("test", datasources = conns)[[1]],
    c("mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb", "new_var", "new_var2")
  )

  expect_equal(
    ds.mean("test$new_var", datasources = conns)$Mean.by.Study[1, 1],
    20090.625
  )

  expect_equal(
    round(ds.mean("test$new_var2", datasources = conns)$Mean.by.Study[1, 1], 2),
    143.09
  )
})

test_that("mutateDS doesn't work with banned function calls that could create a vector", {
  banned_arg_1 <- "banned = row_number()"
  expect_error(
    .check_tidy_disclosure("mtcars", banned_arg_1),
    "`row_number` is not a permitted function")

  banned_arg_2 <- "banned = seq(1, 32, 1)"
  expect_error(
    .check_tidy_disclosure("mtcars", banned_arg_2),
    "`seq` is not a permitted function")

  banned_arg_3 <- "banned = rep(1, 32)"
  expect_error(
    .check_tidy_disclosure("mtcars", banned_arg_3),
    "`rep` is not a permitted function")
})

test_that(".check_mutate_disclosure blocks the use of ':'", {
  banned_arg_4 <- "banned = 1:32"
  expect_error(
    .check_mutate_disclosure(banned_arg_4),
    "It is not permitted to use the character ':' within ds.mutate."
  )
})

test_that(".check_mutate_disclosure blocks the use of 'c'", {
  banned_arg_5 <- "banned = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32)"
  expect_error(
    .check_mutate_disclosure(banned_arg_5),
    "It is not permitted to use the character 'c\\(' within ds.mutate")
})

test_that(".check_mutate_disclosure blocks the use of both 'c' and ':'", {
  banned_arg_6 <- "banned = c(1, 2, 3) + 1:32"
  expect_error(
    .check_mutate_disclosure(banned_arg_6),
    "It is not permitted to use the characters 'c\\( and :' within ds.mutate")
})

test_that("mutateDS permits as.Date on a character column", {
  as_date_arg <- "date_col = as.Date(date_str)"
  as_date_cally <- .make_tidyverse_call("date_df", "mutate", as_date_arg, list(".keep = \"all\", .before = NULL, .after = NULL"))
  result <- eval(as_date_cally)

  expect_equal(
    class(result$date_col),
    "Date"
  )

  expect_equal(
    result$date_col,
    as.Date(c("2024-01-15", "2024-06-30", "2024-12-25"))
  )
})

test_that(".check_mutate_disclosure doesn't block permitted strings", {
  good_arg_1 <- "ok^2"
  expect_silent(
    .check_mutate_disclosure(good_arg_1)
  )
})

test_that(".check_mutate_disclosure doesn't false-positive on function names ending in c", {
  expect_silent(.check_mutate_disclosure("new_col = as.numeric(mpg)"))
  expect_silent(.check_mutate_disclosure("new_col = as.numeric(as.Date(date_str))"))
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
mtcars_group <- mtcars %>% group_by(cyl)
login_data <- .prepare_dslite("renameDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(logins = login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

good_rename_arg <- "test_1 = mpg, test_2 = drat"

test_that("renameDS passes for rename where data and column exist", {
  good_rename_cally <- .make_tidyverse_call("mtcars", "rename", good_rename_arg)
  expected <- c("test_1", "cyl", "disp", "hp", "test_2", "wt", "qsec", "vs", "am", "gear", "carb")
  expect_equal(
    colnames(eval(good_rename_cally)),
    expected
  )
})

test_that("renameDS fails when data doesn't exist", {
  no_data <- .make_tidyverse_call("doesntexist", "rename", good_rename_arg)
  expect_error(
    eval(no_data),
    "`renameDS`\\s+returned\\s+the\\s+following\\s+error:|object\\s+'doesntexist'\\s+not\\s+found"
  )
})

test_that("renameDS fails with bad argument", {
  bad_rename_arg <- "test_1 = mpg, test_2 = drat, filter/asd"
  bad_call <- .make_tidyverse_call("mtcars", "rename", bad_rename_arg)
  expect_error(
    eval(bad_call),
    "Can't rename columns that don't exist\\."
  )
})

test_that("renameDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call(
    "renameDS", "new_name_1$SPACE$$EQU$$SPACE$mpg$COMMA$$SPACE$new_name_2$SPACE$$EQU$$SPACE$drat",
    "mtcars"
  )

  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.colnames("test", datasources = conns)[[1]],
    c("new_name_1", "cyl", "disp", "hp", "new_name_2", "wt", "qsec", "vs", "am", "gear", "carb")
  )
})

require(DSI)
require(DSLite)
require(dplyr)
require(dsBase)
require(dsBaseClient)

data("mtcars")
login_data <- .prepare_dslite("bindColsDS", NULL, list(mtcars = mtcars))
conns <- datashield.login(login_data)
datashield.assign.table(conns, "mtcars", "mtcars")

test_that("bindColsDS passes", {
  bind_cols_arg <- "mtcars, mtcars"
  other_args <- '.name_repair = c("unique", "universal", "check_unique", "minimal")'
  bind_cols_cally <- .make_tidyverse_call(NULL, "bind_cols", bind_cols_arg, other_args, inc_data = F)
  result <- eval(bind_cols_cally)

  expect_equal(
    class(result),
    "data.frame"
  )

  expect_equal(
    colnames(result),
    c(
      "mpg...1", "cyl...2", "disp...3", "hp...4", "drat...5", "wt...6", "qsec...7", "vs...8",
      "am...9", "gear...10", "carb...11", "mpg...12", "cyl...13", "disp...14", "hp...15",
      "drat...16", "wt...17", "qsec...18", "vs...19", "am...20", "gear...21", "carb...22"
    )
  )

  expect_equal(
    dim(result),
    c(32, 22)
  )
})

test_that("bindColsDS passes with .name_repair argument", {
  bind_cols_arg <- "mtcars, mtcars"
  other_args <- '.name_repair = "unique"'
  bind_cols_cally <- .make_tidyverse_call(NULL, "bind_cols", bind_cols_arg, other_args, inc_data = F)
  result <- eval(bind_cols_cally)

  expect_equal(
    colnames(result),
    c(
      "mpg...1", "cyl...2", "disp...3", "hp...4", "drat...5", "wt...6", "qsec...7", "vs...8",
      "am...9", "gear...10", "carb...11", "mpg...12", "cyl...13", "disp...14", "hp...15",
      "drat...16", "wt...17", "qsec...18", "vs...19", "am...20", "gear...21", "carb...22"
    )
  )

  other_args <- '.name_repair = "universal"'
  bind_cols_cally <- .make_tidyverse_call(NULL, "bind_cols", bind_cols_arg, other_args, inc_data = F)
  result <- eval(bind_cols_cally)

  expect_equal(
    colnames(result),
    c(
      "mpg...1", "cyl...2", "disp...3", "hp...4", "drat...5", "wt...6", "qsec...7", "vs...8",
      "am...9", "gear...10", "carb...11", "mpg...12", "cyl...13", "disp...14", "hp...15",
      "drat...16", "wt...17", "qsec...18", "vs...19", "am...20", "gear...21", "carb...22"
    )
  )

  other_args <- '.name_repair = "minimal"'
  bind_cols_cally <- .make_tidyverse_call(NULL, "bind_cols", bind_cols_arg, other_args, inc_data = F)
  result <- eval(bind_cols_cally)

  expect_equal(
    colnames(result),
    c(
      "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb",
      "mpg", "cyl", "disp", "hp", "drat", "wt", "qsec", "vs", "am", "gear", "carb"
    )
  )
})


test_that("bindColsDS fails with .name_repair argument as 'check_unique'", {
  bind_cols_arg <- "mtcars, mtcars"
  other_args <- '.name_repair = "check_unique"'
  bind_cols_cally <- .make_tidyverse_call(NULL, "bind_cols", bind_cols_arg, other_args, inc_data = F)

  expect_error(
    eval(bind_cols_cally),
    "Names must be unique."
  )
})

test_that("bindColsDS passes when called directly", {
  skip_if_not_installed("dsBaseClient")
  cally <- call("bindColsDS", "mtcars$COMMA$$SPACE$mtcars", c("unique", "universal", "check_unique", "minimal"))
  datashield.assign(conns, "test", cally)

  expect_equal(
    ds.class("test", datasources = conns)[[1]],
    "data.frame"
  )

  expect_equal(
    ds.colnames("test", datasources = conns)[[1]],
    c(
      "mpg...1", "cyl...2", "disp...3", "hp...4", "drat...5", "wt...6", "qsec...7", "vs...8",
      "am...9", "gear...10", "carb...11", "mpg...12", "cyl...13", "disp...14", "hp...15",
      "drat...16", "wt...17", "qsec...18", "vs...19", "am...20", "gear...21", "carb...22"
    )
  )

  expect_equal(
    ds.dim("test", datasources = conns)[[1]],
    c(32, 22)
  )
})

#' @title Group by one or more variables
#' @description DataSHIELD implementation of \code{dplyr::group_by}.
#' @param tidy_expr Diffused grouping expression.
#' @param df.name A data frame, data frame extension (e.g. a tibble), or a lazy data frame
#' (e.g. from dbplyr or dtplyr).
#' @param .add When FALSE, the default, \code{group_by()} will override existing groups. To add to
#' the existing groups, use .add = TRUE.
#' @param .drop Drop groups formed by factor levels that don't appear in the data? The default is
#' TRUE except when df.name has been previously grouped with .drop = FALSE.
#' @return A grouped data frame with class grouped_df, unless the combination of \code{tidy_expr}
#' and \code{.add} yields a empty set of grouping columns, in which case a tibble will be returned.
#' @export
groupByDS <- function(tidy_expr, df.name, .add, .drop) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  .check_tidy_disclosure(NULL, tidy_expr, check_df = F)
  other_args <- .paste_character_args(.add, .drop)
  call <- .make_tidyverse_call(df.name, "group_by", tidy_expr, other_args)
  out <- .execute_with_error_handling("group_by", call)
  return(out)
}

#' @title Remove grouping from a tibble or data frame
#' @description DataSHIELD implementation of \code{dplyr::ungroup}.
#' @param tidy_expr Unused in this function.
#' @param x A tibble.
#' @return An ungrouped data frame or tibble.
#' @export
ungroupDS <- function(tidy_expr, x) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  .check_data_name_length(x, listDisclosureSettingsDS())
  call <- .make_tidyverse_call(x, "ungroup", tidy_expr = NULL, other_args = NULL)
  out <- .execute_with_error_handling("ungroup", call)
  return(out)
}

#' @title Coerce a data frame or matrix to a tibble
#' @description DataSHIELD implementation of \code{tibble::as_tibble}. Currently only implemented for data frames and matrices.
#' @param tidy_expr Unused in present function.
#' @param x A data frame or matrix.
#' @param .rows The number of rows, useful to create a 0-column tibble or just as an additional
#' check.
#' @param .name_repair Treatment of problematic column names:
#'   \itemize{
#'     \item "minimal": No name repair or checks, beyond basic existence.
#'     \item "unique": Make sure names are unique and not empty.
#'     \item "check_unique": (default value), no name repair, but check they are unique.
#'     \item "universal": Make the names unique and syntactic.
#'   }
#' @param rownames How to treat existing row names of a data frame or matrix:
#'   \itemize{
#'     \item `NULL`: remove row names. This is the default.
#'     \item `NA`: keep row names.
#'     \item A string: the name of a new column. Existing rownames are transferred
#'     into this column and the \code{row.names} attribute is deleted. No name repair is applied
#'     to the new column name, even if \code{x} already contains a column of that name.
#'   }
#' @return A tibble.
#' @export
asTibbleDS <- function(tidy_expr, x, .rows, .name_repair, rownames) {
  other_args <- .paste_character_args(.rows, .name_repair, rownames)
  call <- .make_tidyverse_call(x, "as_tibble", other_args)
  out <- .execute_with_error_handling("as_tibble", call)
  .check_subset_disclosure_risk(eval(parse(text = x), envir = parent.frame()), out)
  return(out)
}

#' @title Rename columns
#' @description DataSHIELD implementation of \code{dplyr::rename}.
#' @param df.name A data frame or tibble.
#' @param tidy_expr list containing diffused expression.
#' @return An object of the same type as \code{df.name}, typically a data frame or tibble.
#' @export
renameDS <- function(tidy_expr, df.name) {
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  .check_tidy_disclosure(df.name, tidy_expr)
  call <- .make_tidyverse_call(df.name, "rename", tidy_expr)
  out <- .execute_with_error_handling("rename", call)
  return(out)
}

#' @title Create, modify, and delete columns
#' @description DataSHIELD implementation of \code{mutate}.
#' @param df.name A data frame, data frame extension (e.g. a tibble), or a lazy data frame (e.g. from dbplyr or dtplyr).
#' @param tidy_expr Name-value pairs. The name gives the name of the column in the output.
#' @param .keep .keep Control which columns from \code{df.name} are retained in the output. Grouping
#' columns and columns created by \code{tidy_expr} are always kept.
#' \itemize{
#'   \item \code{"all"}: Retains all columns from \code{df.name}. This is the default.
#'   \item \code{"used"}: Retains only the columns used in \code{tidy_expr} to create new columns.
#'   \item \code{"unused"}: Retains only the columns not used in \code{tidy_expr} to create new columns. This is useful if you generate new columns but no longer need the columns used to generate them.
#'   \item \code{"none"}: Doesn't retain any extra columns from \code{df.name}. Only the grouping variables and columns created by \code{tidy_expr} are kept.
#' }
#' Grouping columns and columns created by \code{tidy_expr} are always kept.
#' @param .before <tidy-select> Optionally, control where new columns should appear (the default is
#' to add to the right hand side). See \code{relocate} for more details.
#' @param .after <tidy-select> Optionally, control where new columns should appear (the default is
#' to add to the right hand side). See \code{relocate} for more details.
#' @return An object of the same type as \code{df.name}, typically a data frame or tibble.
#' @export
mutateDS <- function(tidy_expr, df.name, .keep = NULL, .before = NULL, .after = NULL) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  .check_tidy_disclosure(df.name, tidy_expr)
  .check_mutate_disclosure(tidy_expr)
  other_args <- .paste_character_args(.keep, .before, .after)
  call <- .make_tidyverse_call(df.name, "mutate", tidy_expr, other_args)
  out <- .execute_with_error_handling("mutate", call)
  return(out)
}

#' @title Check for Disallowed Character in ds.mutate
#' @description This internal function checks whether the provided expression contains a colon (`:`),
#' which is not permitted in `ds.mutate`. If a colon or letter 'c' is found, the function throws an error.
#' @param tidy_expr A character string representing the expression to be checked.
#' @return This function does not return a value; it throws an error if `tidy_expr` contains `:`.
#' @importFrom cli cli_abort
#' @keywords internal
#' @noRd
.check_mutate_disclosure <- function(tidy_expr){
  matches <- regmatches(tidy_expr, gregexpr("(:|(?<![A-Za-z0-9_.])c\\()", tidy_expr, perl = TRUE))[[1]]
  if (length(matches) > 0) {
    cli_abort(
      c(
        "x" = "It is not permitted to use the character{?s} '{matches}' within ds.mutate.",
        "i" = "You passed the expression '{tidy_expr}'.",
        "i" = "Please remove this character and try again."
      )
    )
  }
}

#' @title Vectorised if-else
#' @description DataSHIELD implementation of \code{dplyr::if_else}.
#' @param condition A list, specifying a logical vector in tidyverse syntax, ie data and column names unquoted.
#' @param true Vector to use for TRUE value of condition.
#' @param false Vector to use for FALSE value of condition.
#' @param missing If not NULL, will be used as the value for NA values of condition. Follows the same size and type rules as true and false.
#' @param ptype An optional prototype declaring the desired output type. If supplied, this overrides the common type of true, false, and missing.
#' @param size An optional size declaring the desired output size. If supplied, this overrides the size of condition.
#' @return A vector with the same size as \code{condition} and the same type as the common type of \code{true}, \code{false}, and \code{missing}.
#' @export
ifElseDS <- function(condition = NULL, true = NULL, false = NULL, missing = NULL,
                     ptype = NULL, size = NULL) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  tidyselect <- .decode_tidy_eval(condition, .get_encode_dictionary())
  .check_tidy_disclosure(NULL, tidyselect, check_df = F)
  other_args <- .paste_character_args(true, false, missing, ptype, size)
  call <- .make_tidyverse_call(.data = NULL, "if_else", tidyselect, other_args, inc_data = F)
  out <- .execute_with_error_handling("if_else", call)
  return(out)
}

#' @title Subset rows using their positions
#' @description DataSHIELD implementation of \code{dplyr::slice}.
#' @param tidy_expr Provide either positive values to keep, or negative values to drop. The values
#' provided must be either all positive or all negative. Indices beyond the number of rows in the
#' input are silently ignored.
#' @param df.name A data frame, data frame extension (e.g. a tibble), or a lazy data frame
#' (e.g. from dbplyr or dtplyr).
#' @param .by Optionally, a selection of columns to group by for just this operation, functioning as
#' an alternative to \code{group_by}.
#' @param .preserve Relevant when the df.name input is grouped. If .preserve = FALSE (the default),
#' the grouping structure is recalculated based on the resulting data, otherwise the grouping is
#' kept as is.
#' @return An object of the same type as \code{df.name}, typically a data frame or tibble.
#' @export
sliceDS <- function(tidy_expr, df.name, .by, .preserve) {
  checkPermissivePrivacyControlLevel(c('permissive', 'banana'))
  tidy_expr <- .decode_tidy_eval(tidy_expr, .get_encode_dictionary())
  .check_tidy_disclosure(df.name, tidy_expr)
  other_args <- .paste_character_args(.by, .preserve)
  call <- .make_tidyverse_call(df.name, "slice", tidy_expr, other_args)
  out <- .execute_with_error_handling("slice", call)
  .check_subset_disclosure_risk(eval(parse(text = df.name), envir = parent.frame()), out)
  return(out)
}

---
title: "Privacy settings within dsTidyverse"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Privacy settings within dsTidyverse}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Data owners can manage privacy settings in two main ways: (i) setting the privacy control level, and (ii) controlling which additional functions can be passed to `ds.tidyverse`.

## Privacy control levels
DataSHIELD implements [privacy control levels](https://wiki.datashield.org/en/opmanag/privacy-control-level), which allows data owners to control which functions can be
used by researchers. The table below shows which dsTidyverse functions are permitted in which
privacy mode. This option can be set on the server: for example to set to non-permissive mode use `datashield.privacyControlLevel = "non-permissive"`

| **Function**       | **Permissive** | **Banana** | **Avocado** | **Non-Permissive** |
|------------------|-------------|---------|---------|---------------|
| `arrangeDS` | ✔ | ✔ |  |  |
| `asTibbleDS` | ✔ | ✔ | ✔  | ✔ |
| `bindColsDS` | ✔ | ✔ |  |  |
| `bindRowsDS` | ✔ | ✔ |  |  |
| `caseWhenDS` | ✔ | ✔ |  |  |
| `distinctDS` | ✔ | ✔ | ✔  | ✔ |
| `filterDS` | ✔ | ✔ |  |  |
| `groupByDS` | ✔ | ✔ |  |  |
| `groupKeysDS` | ✔ | ✔ |  |  |
| `mutateDS` | ✔ | ✔  |  |  |
| `renameDS` | ✔ | ✔ | ✔  | ✔ |
| `selectDS` | ✔ | ✔ | ✔  | ✔ |
| `sliceDS` | ✔ | ✔ |  |  |
| `ungroupDS` | ✔ | ✔ |  |  |

## Permitted functions
dsTidyverse allows additional functions to be passed via the `tidy_expr` argument. For example,
using ds.mutate you can pass `as.numeric`:
```
ds.mutate("mtcars", list(cyl = as.numeric(cyl)), "newobj")
```
Functions are only allowed to be passed which do not risk disclosing individual level data. The default
list of allowed functions is:
```
"everything", "last_col", "group_cols", "starts_with", "ends_with", "contains",
"matches", "num_range", "all_of", "any_of", "where", "rename", "mutate", "if_else",
"case_when", "mean", "median", "mode", "desc", "last_col", "nth", "where", "num_range",
"exp", "sqrt", "scale", "round", "floor", "ceiling", "abs", "sd", "var",
"sin", "cos", "tan", "asin", "acos", "atan", "c", "as.character", "as.integer",
"lag", "diff", "cumsum", "as.Date"
```

These defaults can be managed by the data owner on their server using the option
`tidyverse.permitted.functions`. For example, if as a data owner you want to restrict the permitted
functions to `as.numeric` and `as.integer`, you can set the option
`tidyverse.permitted.functions = c("as.numeric", "as.integer")`.

**WARNING:** This feature gives
the data manager the option to restrict allowed functions, but also to allow additional functions.
If you choose to allow functions not included in the default list, please take steps to ensure that
they are compatible with your research setting, and in sensitive settings with secure data they do not
risk returning individual level data to the researcher. If you have doubts please contact the maintainers
of dsTidyverse who can discuss the risks with you.

# .check_data_name_length throws an error if length of .data exceeds nfilter.string

    Code
      .check_data_name_length(.data, disc_settings)
    Condition
      Error in `.check_data_name_length()`:
      ! Error: The length of string passed to `df.name` must be less than nfilter.string.
      i The value of nfilter.string is:
      80
      i The length of `df.name` is:
      101

# .check_function_names blocks unpermitted function names

    Code
      .check_function_names(arg_unpermitted)
    Condition
      Error:
      ! Values passed to `expr` may only contain permitted functions.
      i Permitted functions are everything, last_col, group_cols, starts_with, ends_with, contains, matches, num_range, all_of, any_of, where, rename, mutate, if_else, case_when, mean, median, desc, ..., is.na, and as.Date.
      i `filter and slice` are not permitted functions.

# .check_variable_length blocks variables with value greater than than nfilter.string

    Code
      .check_variable_length(large_var, disc_settings)
    Condition
      Error:
      ! Error: The maximum length of columns specified in `tidy_select` must be shorter than nfilter.string.
      i The values of nfilter.string are:
      80
      i aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa is longer than this:

# .tidy_disclosure_checks blocks argument with unpermitted variable length

    Code
      .check_tidy_disclosure("dataframe", arg_unpermitted_2)
    Condition
      Error:
      ! Error: The maximum length of columns specified in `tidy_select` must be shorter than nfilter.string.
      i The values of nfilter.string are:
      80
      i aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaasd is longer than this:

# .tidy_disclosure_checks blocks argument with multiple unpermitted function names

    Code
      .check_tidy_disclosure("dataset", arg_unpermitted_3)
    Condition
      Error:
      ! Values passed to `expr` may only contain permitted functions.
      i Permitted functions are everything, last_col, group_cols, starts_with, ends_with, contains, matches, num_range, all_of, any_of, where, rename, mutate, if_else, case_when, mean, median, desc, ..., is.na, and as.Date.
      i `filter and slice` are not permitted functions.

# .tidy_disclosure_checks blocks argument with single unpermitted function name

    Code
      .check_tidy_disclosure("dataset", arg_unpermitted_4)
    Condition
      Error:
      ! Values passed to `expr` may only contain permitted functions.
      i Permitted functions are everything, last_col, group_cols, starts_with, ends_with, contains, matches, num_range, all_of, any_of, where, rename, mutate, if_else, case_when, mean, median, desc, ..., is.na, and as.Date.
      i `slice` is not a permitted function.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{checkPermissivePrivacyControlLevel}
\alias{checkPermissivePrivacyControlLevel}
\title{checkPermissivePrivacyControlLevel}
\usage{
checkPermissivePrivacyControlLevel(privacyControlLevels)
}
\arguments{
\item{privacyControlLevels}{is a vector of strings which contains the privacy control level names which are permitted by the calling method.}
}
\value{
Returns an error if the method is not permitted; otherwise, no value is returned.
}
\description{
This server-side function check that the server is running in "permissive" privacy control level.
}
\details{
Tests whether the R option "datashield.privacyControlLevel" is set to "permissive", if it isn't
will cause a call to stop() with the message "BLOCKED: The server is running in 'non-permissive' mode which
has caused this method to be blocked".
}
\author{
Wheater, Dr SM., DataSHIELD Team.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renameDS.R
\name{renameDS}
\alias{renameDS}
\title{Rename columns}
\usage{
renameDS(tidy_expr, df.name)
}
\arguments{
\item{tidy_expr}{list containing diffused expression.}

\item{df.name}{A data frame or tibble.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::rename}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caseWhenDS.R
\name{caseWhenDS}
\alias{caseWhenDS}
\title{Performs dplyr case_when}
\usage{
caseWhenDS(tidy_expr = NULL, .default = NULL, .ptype = NULL, .size = NULL)
}
\arguments{
\item{tidy_expr}{A sequence of two-sided formulas. The left hand side (LHS) determines which
values match this case. The right hand side (RHS) provides the replacement value.
The LHS inputs must evaluate to logical vectors.
The RHS inputs will be coerced to their common type.
All inputs will be recycled to their common size. That said, we encourage all LHS inputs to be
the same size. Recycling is mainly useful for RHS inputs, where you might supply a size 1 input
that will be recycled to the size of the LHS inputs.
NULL inputs are ignored.}

\item{.default}{The value used when all of the LHS inputs return either FALSE or NA.}

\item{.ptype}{An optional prototype declaring the desired output type. If supplied, this overrides the common type of true, false, and missing.}

\item{.size}{An optional size declaring the desired output size. If supplied, this overrides the size of condition.}
}
\value{
A vector with the same size as the common size computed from the inputs in \code{tidy_expr} and the same type as the common type of the RHS inputs in \code{tidy_expr}.
}
\description{
DataSHIELD implementation of \code{dplyr::case_when}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutateDS.R
\name{mutateDS}
\alias{mutateDS}
\title{Create, modify, and delete columns}
\usage{
mutateDS(tidy_expr, df.name, .keep = NULL, .before = NULL, .after = NULL)
}
\arguments{
\item{tidy_expr}{Name-value pairs. The name gives the name of the column in the output.}

\item{df.name}{A data frame, data frame extension (e.g. a tibble), or a lazy data frame (e.g. from dbplyr or dtplyr).}

\item{.keep}{.keep Control which columns from \code{df.name} are retained in the output. Grouping
columns and columns created by \code{tidy_expr} are always kept.
\itemize{
  \item \code{"all"}: Retains all columns from \code{df.name}. This is the default.
  \item \code{"used"}: Retains only the columns used in \code{tidy_expr} to create new columns.
  \item \code{"unused"}: Retains only the columns not used in \code{tidy_expr} to create new columns. This is useful if you generate new columns but no longer need the columns used to generate them.
  \item \code{"none"}: Doesn't retain any extra columns from \code{df.name}. Only the grouping variables and columns created by \code{tidy_expr} are kept.
}
Grouping columns and columns created by \code{tidy_expr} are always kept.}

\item{.before}{<tidy-select> Optionally, control where new columns should appear (the default is
to add to the right hand side). See \code{relocate} for more details.}

\item{.after}{<tidy-select> Optionally, control where new columns should appear (the default is
to add to the right hand side). See \code{relocate} for more details.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{mutate}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bindColsDS.R
\name{bindColsDS}
\alias{bindColsDS}
\title{Bind multiple data frames by column}
\usage{
bindColsDS(to_combine = NULL, .name_repair = NULL)
}
\arguments{
\item{to_combine}{Data frames to combine. Each argument can either be a data frame, a list that
could be a data frame, or a list of data frames. Columns are matched by name, and any missing
columns will be filled with NA.}

\item{.name_repair}{One of "unique", "universal", or "check_unique". See
\code{vctrs::vec_as_names()} for the meaning of these options.}
}
\value{
A data frame the same type as the first element of \code{to_combine}
}
\description{
DataSHIELD implementation of \code{dplyr::bind_cols}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterDS.R
\name{filterDS}
\alias{filterDS}
\title{Performs dplyr filter}
\usage{
filterDS(tidy_expr, df.name, .by, .preserve)
}
\arguments{
\item{tidy_expr}{Diffused expression that return a logical value, and are defined in terms of the
variables in \code{df.name}.}

\item{df.name}{A data frame, data frame extension (e.g. a tibble), or a lazy data frame
(e.g. from dbplyr or dtplyr).}

\item{.by}{Optionally, a selection of columns to group by for just this operation, functioning as
an alternative to \code{group_by}.}

\item{.preserve}{Relevant when the df.name input is grouped. If .preserve = FALSE (the default),
the grouping structure is recalculated based on the resulting data, otherwise the grouping is kept as is.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::filter}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{listPermittedTidyverseFunctionsDS}
\alias{listPermittedTidyverseFunctionsDS}
\title{List of Permitted Tidyverse Functions}
\usage{
listPermittedTidyverseFunctionsDS()
}
\value{
A character vector of function names, each representing a permitted function. Functions
not included in this list will be blocked.
}
\description{
This function returns a vector of function names that are permitted to be passed within the
dsTidyverse functions, e.g. within the `tidy_select` argument of `ds.mutate.`
Additional functions can be added via the \code{tidyverse.permitted.functions} R option.
Only functions from base R, dplyr, or tidyselect are supported.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distinctDS.R
\name{distinctDS}
\alias{distinctDS}
\title{Keep distinct/unique rows}
\usage{
distinctDS(tidy_expr, df.name, .keep_all)
}
\arguments{
\item{tidy_expr}{Optional variables to use when determining uniqueness. If there are multiple rows for
a given combination of inputs, only the first row will be preserved. If omitted, will use all
variables in the data frame.}

\item{df.name}{A data frame, data frame extension (e.g. a tibble), or a lazy data frame
(e.g. from dbplyr or dtplyr).}

\item{.keep_all}{If TRUE, keep all variables in df.name If a combination of expr is not distinct,
this keeps the first row of values.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::distinct}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sliceDS.R
\name{sliceDS}
\alias{sliceDS}
\title{Subset rows using their positions}
\usage{
sliceDS(tidy_expr, df.name, .by, .preserve)
}
\arguments{
\item{tidy_expr}{Provide either positive values to keep, or negative values to drop. The values
provided must be either all positive or all negative. Indices beyond the number of rows in the
input are silently ignored.}

\item{df.name}{A data frame, data frame extension (e.g. a tibble), or a lazy data frame
(e.g. from dbplyr or dtplyr).}

\item{.by}{Optionally, a selection of columns to group by for just this operation, functioning as
an alternative to \code{group_by}.}

\item{.preserve}{Relevant when the df.name input is grouped. If .preserve = FALSE (the default),
the grouping structure is recalculated based on the resulting data, otherwise the grouping is
kept as is.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::slice}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/selectDS.R
\name{selectDS}
\alias{selectDS}
\title{Keep or drop columns using their names and types}
\usage{
selectDS(tidy_expr, df.name)
}
\arguments{
\item{tidy_expr}{One or more unquoted expressions separated by commas.}

\item{df.name}{A data frame or tibble.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::select}.
}
\details{
Performs dplyr select
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groupByDS.R
\name{groupByDS}
\alias{groupByDS}
\title{Group by one or more variables}
\usage{
groupByDS(tidy_expr, df.name, .add, .drop)
}
\arguments{
\item{tidy_expr}{Diffused grouping expression.}

\item{df.name}{A data frame, data frame extension (e.g. a tibble), or a lazy data frame
(e.g. from dbplyr or dtplyr).}

\item{.add}{When FALSE, the default, \code{group_by()} will override existing groups. To add to
the existing groups, use .add = TRUE.}

\item{.drop}{Drop groups formed by factor levels that don't appear in the data? The default is
TRUE except when df.name has been previously grouped with .drop = FALSE.}
}
\value{
A grouped data frame with class grouped_df, unless the combination of \code{tidy_expr}
and \code{.add} yields a empty set of grouping columns, in which case a tibble will be returned.
}
\description{
DataSHIELD implementation of \code{dplyr::group_by}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groupKeysDS.R
\name{groupKeysDS}
\alias{groupKeysDS}
\title{Performs dplyr \code{group_keys}.}
\usage{
groupKeysDS(tidy_select, x)
}
\arguments{
\item{tidy_select}{Unused in this function.}

\item{x}{a grouped tibble.}
}
\value{
A data frame describing the groups.
}
\description{
DataSHIELD implementation of \code{dplyr::group_keys}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/asTibbleDS.R
\name{asTibbleDS}
\alias{asTibbleDS}
\title{Coerce a data frame or matrix to a tibble}
\usage{
asTibbleDS(tidy_expr, x, .rows, .name_repair, rownames)
}
\arguments{
\item{tidy_expr}{Unused in present function.}

\item{x}{A data frame or matrix.}

\item{.rows}{The number of rows, useful to create a 0-column tibble or just as an additional
check.}

\item{.name_repair}{Treatment of problematic column names:
\itemize{
  \item "minimal": No name repair or checks, beyond basic existence.
  \item "unique": Make sure names are unique and not empty.
  \item "check_unique": (default value), no name repair, but check they are unique.
  \item "universal": Make the names unique and syntactic.
}}

\item{rownames}{How to treat existing row names of a data frame or matrix:
\itemize{
  \item `NULL`: remove row names. This is the default.
  \item `NA`: keep row names.
  \item A string: the name of a new column. Existing rownames are transferred
  into this column and the \code{row.names} attribute is deleted. No name repair is applied
  to the new column name, even if \code{x} already contains a column of that name.
}}
}
\value{
A tibble.
}
\description{
DataSHIELD implementation of \code{tibble::as_tibble}. Currently only implemented for data frames and matrices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bindRowsDS.R
\name{bindRowsDS}
\alias{bindRowsDS}
\title{Bind multiple data frames by row.}
\usage{
bindRowsDS(to_combine = NULL, .id = NULL)
}
\arguments{
\item{to_combine}{Data frames to combine. Each argument can either be a data frame, a list that
could be a data frame, or a list of data frames. Columns are matched by name, and any missing
columns will be filled with NA.}

\item{.id}{he name of an optional identifier column. Provide a string to create an output column
that identifies each input. The column will use names if available, otherwise it will use
positions.}
}
\value{
A data frame the same type as the first element of \code{to_combine}
}
\description{
DataSHIELD implementation of \code{dplyr::bind_rows}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/groupByDS.R
\name{ungroupDS}
\alias{ungroupDS}
\title{Remove grouping from a tibble or data frame}
\usage{
ungroupDS(tidy_expr, x)
}
\arguments{
\item{tidy_expr}{Unused in this function.}

\item{x}{A tibble.}
}
\value{
An ungrouped data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::ungroup}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arrangeDS.R
\name{arrangeDS}
\alias{arrangeDS}
\title{Order the rows of a data frame by the values of selected columns}
\usage{
arrangeDS(tidy_expr, df.name, .by_group)
}
\arguments{
\item{tidy_expr}{Variables, or functions of variables. Use \code{desc()} to sort a variable in descending
order.}

\item{df.name}{A data frame, data frame extension (e.g. a tibble), or a lazy data frame (e.g. from dbplyr or dtplyr).}

\item{.by_group}{If TRUE, will sort first by grouping variable. Applies to grouped data frames
only.}
}
\value{
An object of the same type as \code{df.name}, typically a data frame or tibble.
}
\description{
DataSHIELD implementation of \code{dplyr::arrange}.
}
